__all__ = (
    "MoMetricBundle",
    "MoMetricBundleGroup",
    "create_empty_mo_metric_bundle",
    "make_completeness_bundle",
)

import os
import warnings

import matplotlib.pyplot as plt
import numpy as np
import numpy.ma as ma

from rubin_sim.maf.metrics import BaseMoMetric, ValueAtHMetric
from rubin_sim.maf.plots import PlotHandler
from rubin_sim.maf.slicers import MoObjSlicer
from rubin_sim.maf.stackers import BaseMoStacker, MoMagStacker

from .metric_bundle import MetricBundle


def create_empty_mo_metric_bundle():
    """Create an empty metric bundle.

    Returns
    -------
    MoMetricBundle : `~rubin_sim.maf.metricBundles.MoMetricBundle`
        An empty metric bundle, configured with just
        the :class:`BaseMetric` and :class:`BaseSlicer`.
    """
    return MoMetricBundle(BaseMoMetric(), MoObjSlicer(), None)


def make_completeness_bundle(bundle, completeness_metric, h_mark=None, results_db=None):
    """Evaluate a MoMetricBundle with a completeness-style metric, and
    downsample into a new MoMetricBundle marginalized over the population.

    Parameters
    ----------
    bundle : `~rubin_sim.maf.metricBundles.MoMetricBundle`
        The metric bundle with a completeness summary statistic.
    completeness_metric : `~rubin_sim.maf.metric`
        The summary (completeness) metric to run on the bundle.
    h_mark : `float`, optional
        The Hmark value to add to the plotting dictionary of the new
        mock bundle. Default None.
    results_db : `~rubin_sim.maf.db.ResultsDb`, optional
        The results_db in which to record the summary statistic value at
        Hmark. Default None.

    Returns
    -------
    mo_metric_bundle : `~rubin_sim.maf.metricBundles.MoMetricBundle`

    Notes
    -----
    This utility turns a metric bundle which could evaluate a metric over
    the population, into a secondary or mock metric bundle, using either
    MoCompleteness or MoCumulativeCompleteness summary
    metrics to marginalize over the population of moving objects.
    This lets us use the plotHandler + plots.MetricVsH
    to generate plots across the population, using the completeness
    information.
    This utility will also work with completeness metric run in order
    to calculate fraction of the population,
    or with MoCompletenessAtTime metric.
    """
    bundle.set_summary_metrics(completeness_metric)
    # This step adds summary values at each point to the original metric -
    # we use this to populate the completeness values in the next step.
    # However, we may not want them to go into the results_db.
    bundle.compute_summary_stats(results_db)
    summary_name = completeness_metric.name
    # Make up the bundle, including the metric values.
    completeness = ma.MaskedArray(
        data=bundle.summary_values[summary_name]["value"],
        mask=np.zeros(len(bundle.summary_values[summary_name]["value"])),
        fill_value=0,
    )
    mb = MoMetricBundle(
        completeness_metric,
        bundle.slicer,
        constraint=bundle.constraint,
        run_name=bundle.run_name,
        info_label=bundle.info_label,
        display_dict=bundle.display_dict,
    )
    plot_dict = {}
    plot_dict.update(bundle.plot_dict)
    plot_dict["label"] = bundle.info_label
    if "Completeness" not in summary_name:
        plot_dict["label"] += " " + summary_name.replace("FractionPop_", "")
    mb.metric_values = completeness.reshape(1, len(completeness))
    if h_mark is not None:
        metric = ValueAtHMetric(h_mark=h_mark)
        mb.set_summary_metrics(metric)
        mb.compute_summary_stats(results_db)
        val = mb.summary_values["Value At H=%.1f" % h_mark]
        if val is None:
            val = 0
        if summary_name.startswith("Cumulative"):
            plot_dict["label"] += ": @ H(<=%.1f) = %.1f%s" % (h_mark, val * 100, "%")
        else:
            plot_dict["label"] += ": @ H(=%.1f) = %.1f%s" % (h_mark, val * 100, "%")
    mb.set_plot_dict(plot_dict)
    return mb


class MoMetricBundle(MetricBundle):
    """Define a moving object metric bundle combination of
    moving-object metric, moving-object slicer, and constraint.

    Parameters
    ----------
    metric : `~rubin_sim.maf.metric`
        The Metric class to run per slice_point
    slicer : `~rubin_sim.maf.slicer`
        The Slicer to apply to the incoming visit data (the observations).
    constraint : `str` or None, opt
        A (sql-style) constraint to apply to the visit data, to apply a
        broad sub-selection.
    stacker_list : `list` [`~rubin_sim.maf.stacker`], opt
        A list of pre-configured stackers to use to generate additional
        columns per visit.
        These will be generated automatically if needed, but pre-configured
        versions will override these.
    run_name : `str`, opt
        The name of the simulation being run.
        This will be added to output files and plots.
        Setting it prevents file conflicts when running the same
        metric on multiple simulations, and
        provides a way to identify which simulation is being analyzed.
    info_label : `str` or None, opt
        Information to add to the output metric data file name and plot labels.
        If this is not provided, it will be auto-generated from the
        constraint (if any).
        Setting this provides an easy way to specify different
        configurations of a metric, a slicer,
        or just to rewrite your constraint into friendlier terms.
        (i.e. a constraint like 'note not like "%DD%"' can become
        "non-DD" in the file name and plot labels
        by specifying info_label).
    plot_dict : `dict` of plotting parameters, opt
        Specify general plotting parameters, such as x/y/color limits.
    display_dict : `dict` of display parameters, opt
        Specify parameters for show_maf web pages, such as the
        side bar labels and figure captions.
        Keys: 'group', 'subgroup', 'caption', and 'order'
        (such as to set metrics in filter order, etc)
    child_metrics : `list` of `~rubin_sim.maf.metrics`
        A list of child metrics to run to summarize the
        primary metric, such as Discovery_At_Time, etc.
    summary_metrics : `list` of `~rubin_sim.maf.metrics`
        A list of summary metrics to run to summarize the
        primary or child metric, such as CompletenessAtH, etc.

    Notes
    -----
    Define the "thing" you are measuring, with a combination of
    * metric (calculated per object)
    * slicer (contains information on the moving objects
    and their observations)
    * constraint (an optional definition of a large subset of data)

    The MoMetricBundle also saves the child metrics to be used
    to generate summary statistics over those metric values,
    as well as the resulting summary statistic values.

    Plotting parameters and display parameters (for show_maf) are saved
    in the MoMetricBundle, as well as additional info_label such as the
    opsim run name, and relevant stackers and maps
    to apply when calculating the metric values.
    """

    def __init__(
        self,
        metric,
        slicer,
        constraint=None,
        stacker_list=None,
        run_name="run name",
        info_label=None,
        file_root=None,
        plot_dict=None,
        plot_funcs=None,
        display_dict=None,
        child_metrics=None,
        summary_metrics=None,
    ):
        self.metric = metric
        self.slicer = slicer
        if constraint == "":
            constraint = None
        self.constraint = constraint
        # Set the stackerlist.
        if stacker_list is not None:
            if isinstance(stacker_list, BaseMoStacker):
                self.stacker_list = [
                    stacker_list,
                ]
            else:
                self.stacker_list = []
                for s in stacker_list:
                    if not isinstance(s, BaseMoStacker):
                        raise ValueError(
                            "stackerList must only contain " "rubin_sim.maf.stackers.BaseMoStacker type objs"
                        )
                    self.stacker_list.append(s)
        else:
            self.stacker_list = []
        # Add the basic 'visibility/mag' stacker if not present.
        mag_stacker_found = False
        for s in self.stacker_list:
            if s.__class__.__name__ == "MoMagStacker":
                mag_stacker_found = True
                break
        if not mag_stacker_found:
            self.stacker_list.append(MoMagStacker())
        # Set a mapsList just for compatibility with generic MetricBundle.
        self.maps_list = []
        # Add the summary stats, if applicable.
        self.set_summary_metrics(summary_metrics)
        # Set the provenance/info_label.
        self.run_name = run_name
        self._build_metadata(info_label)
        # Build the output filename root if not provided.
        if file_root is not None:
            self.file_root = file_root
        else:
            self._build_file_root()
        # Set the plotting classes/functions.
        self.set_plot_funcs(plot_funcs)
        # Set the plot_dict and displayDicts.
        self.plot_dict = {"units": "@H"}
        self.set_plot_dict(plot_dict)
        # Update/set display_dict.
        self.display_dict = {}
        self.set_display_dict(display_dict)
        # Set the list of child metrics.
        self.set_child_bundles(child_metrics)
        # This is where we store the metric values and summary stats.
        self.metric_values = None
        self.summary_values = None

    def _reset_metric_bundle(self):
        """Reset all properties of MetricBundle."""
        self.metric = None
        self.slicer = None
        self.constraint = None
        self.stacker_list = [MoMagStacker()]
        self.maps_list = []
        self.summary_metrics = []
        self.plot_funcs = []
        self.run_name = "opsim"
        self.info_label = ""
        self.db_cols = None
        self.file_root = None
        self.plot_dict = {}
        self.display_dict = {}
        self.child_metrics = None
        self.metric_values = None
        self.summary_values = None

    def _build_metadata(self, info_label):
        """If no info_label is provided, auto-generate it from the
        obs_file + constraint."""
        if info_label is None:
            try:
                self.info_label = self.slicer.obsfile.replace(".txt", "").replace(".dat", "")
                self.info_label = self.info_label.replace("_obs", "").replace("_allObs", "")
            except AttributeError:
                self.info_label = "noObs"
            # And modify by constraint.
            if self.constraint is not None:
                self.info_label += " " + self.constraint
        else:
            self.info_label = info_label

    def _find_req_cols(self):
        # Doesn't quite work the same way yet. No stacker list, for example.
        raise NotImplementedError

    def set_child_bundles(self, child_metrics=None):
        """Identify any child metrics to be run on this (parent) bundle.
        and create the new metric bundles that will hold the child values,
        linking to this bundle.
        Remove the summaryMetrics from self afterwards.

        Parameters
        ----------
        child_metrics : `~maf.MoMetric`
            Child metrics work like reduce functions for non-moving objects.
            They pull out subsets of the original metric values, typically
            do more processing on those values, and then save them in
            new metric bundles.
        """
        self.child_bundles = {}
        if child_metrics is None:
            child_metrics = self.metric.child_metrics
        for c_name, cMetric in child_metrics.items():
            c_bundle = MoMetricBundle(
                metric=cMetric,
                slicer=self.slicer,
                constraint=self.constraint,
                stacker_list=self.stacker_list,
                run_name=self.run_name,
                info_label=self.info_label,
                plot_dict=self.plot_dict,
                plot_funcs=self.plot_funcs,
                display_dict=self.display_dict,
                summary_metrics=self.summary_metrics,
            )
            self.child_bundles[c_name] = c_bundle
        if len(child_metrics) > 0:
            self.summary_metrics = []

    def compute_summary_stats(self, results_db=None):
        """Compute summary statistics on metric_values, using summaryMetrics,
        for self and child bundles.

        Parameters
        ----------
        results_db : `~maf.ResultsDb`
            Database which holds the summary statistic information.
        """
        if self.summary_values is None:
            self.summary_values = {}
        if self.summary_metrics is not None:
            # Build array of metric values, to use for summary statistics.
            for m in self.summary_metrics:
                summary_name = m.name
                summary_val = m.run(self.metric_values, self.slicer.slice_points["H"])
                self.summary_values[summary_name] = summary_val
                # Add summary metric info to results database, if applicable.
                if results_db:
                    metric_id = results_db.update_metric(
                        self.metric.name,
                        self.slicer.slicer_name,
                        self.run_name,
                        self.constraint,
                        self.info_label,
                        None,
                    )
                    results_db.update_summary_stat(
                        metric_id, summary_name=summary_name, summary_value=summary_val
                    )

    def reduce_metric(self, reduce_func, reduce_plot_dict=None, reduce_display_dict=None):
        raise NotImplementedError


class MoMetricBundleGroup:
    """Run groups of MoMetricBundles.

    Parameters
    ----------
    bundle_dict : `dict` or `list` [`MoMetricBundles`]
        Individual MoMetricBundles should be placed into a dictionary,
        and then passed to the MoMetricBundleGroup.
        The dictionary keys can then be used to identify MoMetricBundles
        if needed -- and to identify new MetricBundles which could be
        created if 'reduce' functions are run on a particular MoMetricBundle.
        MoMetricBundles must all have the same Slicer (same set of moving
        object observations).
    out_dir : `str`, opt
        Directory to save the metric results.
        Default is the current directory.
    results_db : `ResultsDb`, opt
        A results database to store summary stat information.
        If not specified, one will be created in the out_dir.
        This database saves information about the metrics calculated,
        including their summary statistics.
    verbose : `bool`, opt
        Flag to turn on/off verbose feedback.
    """

    def __init__(self, bundle_dict, out_dir=".", results_db=None, verbose=True):
        self.verbose = verbose
        self.bundle_dict = bundle_dict
        self.out_dir = out_dir
        if not os.path.isdir(self.out_dir):
            os.makedirs(self.out_dir)
        self.results_db = results_db

        self.slicer = list(self.bundle_dict.values())[0].slicer
        for b in self.bundle_dict.values():
            if b.slicer != self.slicer:
                raise ValueError(
                    "Currently, the slicers for the MoMetricBundleGroup must be equal,"
                    " using the same observations and Hvals."
                )
        self.constraints = list(set([b.constraint for b in bundle_dict.values()]))

    def _check_compatible(self, metric_bundle1, metric_bundle2):
        """Check if two MetricBundles are "compatible".
        Compatible indicates that the constraints, the slicers,
        and the maps are the same, and
        that the stackers do not interfere with each other
        (i.e. are not trying to set the same column in different ways).
        Returns True if the MetricBundles are compatible, False if not.

        Parameters
        ----------
        metric_bundle1 : `MetricBundle`
        metric_bundle2 : `MetricBundle`

        Returns
        -------
        match : `bool`
        """
        if metric_bundle1.constraint != metric_bundle2.constraint:
            return False
        if metric_bundle1.slicer != metric_bundle2.slicer:
            return False
        if metric_bundle1.maps_list.sort() != metric_bundle2.maps_list.sort():
            return False
        for stacker in metric_bundle1.stacker_list:
            for stacker2 in metric_bundle2.stacker_list:
                # If the stackers have different names, that's OK,
                # and if they are identical, that's ok.
                if (stacker.__class__.__name__ == stacker2.__class__.__name__) & (stacker != stacker2):
                    return False
        # But if we got this far, everything matches.
        return True

    def _find_compatible(self, test_keys):
        """Private utility to find which metricBundles with keys in the
        list 'test_keys' can be calculated
        at the same time -- having the same slicer, constraint, maps,
        and compatible stackers.

        Parameters
        -----------
        test_keys : `list`
            List of the dictionary keys (of self.bundle_dict) to
            test for compatibility.

        Returns
        --------
        compatible_lists : `list` [`lists`]
            Returns test_keys, split into separate lists of
            compatible metricBundles.
        """
        compatible_lists = []
        for k in test_keys:
            try:
                b = self.bundle_dict[k]
            except KeyError:
                warnings.warn(
                    "Received %s in testkeys, but this is not present in self.bundle_dict."
                    "Will continue, but this is not expected."
                )
                continue
            found_compatible = False
            checked_all = False
            while not (found_compatible) and not (checked_all):
                # Go through the existing lists in compatible_lists, to see
                # if this metricBundle matches.
                for compatible_list in compatible_lists:
                    # Compare to all the metricBundles in this subset,
                    # to check all stackers are compatible.
                    found_compatible = True
                    for comparison_key in compatible_list:
                        compatible = self._check_compatible(self.bundle_dict[comparison_key], b)
                        if not compatible:
                            # Found a metricBundle which is not compatible,
                            # so stop and go onto the next subset.
                            found_compatible = False
                            break
                checked_all = True
            if found_compatible:
                compatible_list.append(k)
            else:
                compatible_lists.append(
                    [
                        k,
                    ]
                )
        return compatible_lists

    def run_constraint(self, constraint):
        """Calculate the metric values for all the metricBundles which
        match this constraint in the metricBundleGroup.
        Also calculates child metrics and summary statistics,
        and writes all to disk.

        Parameters
        ----------
        constraint : `str`
            SQL-where or pandas constraint for the metricBundles.
        """
        # Find the dict keys of the bundles which match this constraint.
        keys_matching_constraint = []
        for k, b in self.bundle_dict.items():
            if b.constraint == constraint:
                keys_matching_constraint.append(k)
        if len(keys_matching_constraint) == 0:
            return
        # Identify the observations which are relevant for this constraint.
        # This sets slicer.obs (valid for all H values).
        self.slicer.subset_obs(constraint)
        # Identify the sets of these metricBundles can be run at the same time
        # (also have the same stackers).
        compatible_lists = self._find_compatible(keys_matching_constraint)

        # And now run each of those subsets of compatible metricBundles.
        for compatible_list in compatible_lists:
            self._run_compatible(compatible_list)

    def _run_compatible(self, compatible_list):
        """Calculate the metric values for set of (parent and child) bundles,
        as well as the summary stats, and write to disk.

        Parameters
        -----------
        compatible_list : `list`
            List of dictionary keys, of the metricBundles which can be
            calculated together. This means they are 'compatible' and have
            the same slicer, constraint, and non-conflicting mappers and
            stackers.
        """
        if self.verbose:
            print("Running metrics %s" % compatible_list)

        b_dict = self.bundle_dict

        # Find the unique stackers and maps.
        # These are already "compatible" (as id'd by compatible_list).
        uniq_stackers = []
        all_stackers = []
        uniq_maps = []
        all_maps = []
        for b in b_dict.values():
            all_stackers += b.stacker_list
            all_maps += b.maps_list
        for s in all_stackers:
            if s not in uniq_stackers:
                uniq_stackers.append(s)
        for m in all_maps:
            if m not in uniq_maps:
                uniq_maps.append(m)

        if len(uniq_maps) > 0:
            print("Got some maps .. that was unexpected at the moment. Can't use them here yet.")

        # Set up all of the metric values, including for the child bundles.
        for k in compatible_list:
            b = self.bundle_dict[k]
            b._setup_metric_values()
            for cb in b.child_bundles.values():
                cb._setup_metric_values()
        # Calculate the metric values.
        for i, slice_point in enumerate(self.slicer):
            sso_obs = slice_point["obs"]
            for j, Hval in enumerate(slice_point["Hvals"]):
                # Run stackers to add extra columns (that depend on h_val)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    for s in uniq_stackers:
                        sso_obs = s.run(sso_obs, slice_point["orbit"]["H"], Hval)
                # Run all the parent metrics.
                for k in compatible_list:
                    b = self.bundle_dict[k]
                    # Mask the parent metric (and then child metrics)
                    # if there was no data.
                    if len(sso_obs) == 0:
                        b.metric_values.mask[i][j] = True
                        for cb in list(b.child_bundles.values()):
                            cb.metric_values.mask[i][j] = True
                    # Otherwise, calculate the metric value for the parent,
                    # and then child.
                    else:
                        # Calculate for the parent.
                        m_val = b.metric.run(sso_obs, slice_point["orbit"], Hval)
                        # Mask if the parent metric returned a bad value.
                        if m_val == b.metric.badval:
                            b.metric_values.mask[i][j] = True
                            for cb in b.child_bundles.values():
                                cb.metric_values.mask[i][j] = True
                        # Otherwise, set the parent value and calculate
                        # the child metric values as well.
                        else:
                            b.metric_values.data[i][j] = m_val
                            for cb in b.child_bundles.values():
                                child_val = cb.metric.run(sso_obs, slice_point["orbit"], Hval, m_val)
                                if child_val == cb.metric.badval:
                                    cb.metric_values.mask[i][j] = True
                                else:
                                    cb.metric_values.data[i][j] = child_val
        for k in compatible_list:
            b = self.bundle_dict[k]
            b.compute_summary_stats(self.results_db)
            for c_b in b.child_bundles.values():
                c_b.compute_summary_stats(self.results_db)
                # Write to disk.
                c_b.write(out_dir=self.out_dir, results_db=self.results_db)
            # Write to disk.
            b.write(out_dir=self.out_dir, results_db=self.results_db)

    def run_all(self):
        """Run all constraints and metrics for these moMetricBundles."""
        for constraint in self.constraints:
            self.run_constraint(constraint)
        if self.verbose:
            print("Calculated and saved all metrics.")

    def plot_all(
        self,
        savefig=True,
        outfile_suffix=None,
        fig_format="pdf",
        dpi=600,
        thumbnail=True,
        closefigs=True,
    ):
        """
        Make a few generically desired plots.
        Given the nature of the outputs for much of the moving object
        metrics, a good deal of the plotting for the moving object batch
        is handled in a custom manner joining together multiple
        metricsbundles.
        """
        plot_handler = PlotHandler(
            out_dir=self.out_dir,
            results_db=self.results_db,
            savefig=savefig,
            fig_format=fig_format,
            dpi=dpi,
            thumbnail=thumbnail,
        )
        for b in self.bundle_dict.values():
            try:
                b.plot(
                    plot_handler=plot_handler,
                    outfile_suffix=outfile_suffix,
                    savefig=savefig,
                )
            except ValueError as ve:
                message = "Plotting failed for metricBundle %s." % (b.file_root)
                message += " Error message: %s" % (ve.message)
                warnings.warn(message)
            if closefigs:
                plt.close("all")
        if self.verbose:
            print("Plotting all metrics.")



__all__ = ("add_run",)

import argparse

from . import add_run_to_database


def add_run():
    parser = argparse.ArgumentParser(description="Add a MAF run to the tracking database.")
    parser.add_argument("maf_dir", type=str, help="Directory containing MAF outputs.")
    parser.add_argument("-c", "--maf_comment", type=str, default=None, help="Comment on MAF analysis.")
    parser.add_argument("--group", type=str, default=None, help="Opsim Group name.")
    parser.add_argument("--run_name", type=str, default=None, help="Run Name.")
    parser.add_argument("--run_comment", type=str, default=None, help="Comment on OpSim run.")
    parser.add_argument("--db_file", type=str, default="None", help="Opsim Sqlite filename")
    defaultdb = "trackingDb_sqlite.db"
    parser.add_argument(
        "-t",
        "--tracking_db",
        type=str,
        default=defaultdb,
        help="Tracking database filename. Default is %s, in the current directory." % defaultdb,
    )
    args = parser.parse_args()

    add_run_to_database(
        maf_dir=args.maf_dir,
        tracking_db_file=args.tracking_db,
        run_group=args.group,
        run_name=args.run_name,
        run_comment=args.run_comment,
        maf_comment=args.maf_comment,
        db_file=args.db_file,
    )



__all__ = ("SNNSNMetric",)


import healpy as hp
import numpy as np
import numpy.lib.recfunctions as nlr
import pandas as pd
from scipy.interpolate import interp1d

from rubin_sim.maf.metrics import BaseMetric
from rubin_sim.maf.utils.sn_n_sn_utils import LcfastNew, SnRate, load_sne_cached
from rubin_sim.phot_utils import DustValues


class SNNSNMetric(BaseMetric):
    """
    Measure zlim of type Ia supernovae.

    Parameters
    --------------
    metricName : `str`, opt
        metric name (default : SNSNRMetric)
    mjd_col : `str`, opt
        mjd column name (default : observationStartMJD)
    filter_col : `str`, opt
        filter column name (default: filter)
    m5_col : `str`, opt
        five-sigma depth column name (default : fiveSigmaDepth)
    exptime_col : `str`, opt
        exposure time column name (default : visitExposureTime)
    night_col : `str`, opt
        night column name (default : night)
    obsid_col : `str`, opt
        observation id column name (default : observationId)
    nexp_col : `str`, opt
        number of exposure column name (default : numExposures)
    vistime_col : `str`, opt
        visit time column name (default : visitTime)
    seeing_col : `str`, opt
        seeing column name (default: seeingFwhmEff)
    note_col : `str`, opt
        note column name (default: note)
    season : `list`, opt
        list of seasons to process (float)(default: -1 = all seasons)
    coadd : `bool`, opt
        coaddition per night (and per band) (default : True)
    zmin : `float`, opt
        min redshift for the study (default: 0.0)
    zmax : `float`, opt
        max redshift for the study (default: 1.2)
    verbose : `bool`, opt
        verbose mode (default: False)
    n_bef : `int`, opt
        number of LC points LC before T0 (default:5)
    n_aft : `int`, opt
        number of LC points after T0 (default: 10)
    snr_min : `float`, opt
        minimal SNR of LC points (default: 5.0)
    n_phase_min : `int`, opt
        number of LC points with phase<= -5(default:1)
    n_phase_max : `int`, opt
        number of LC points with phase>= 20 (default: 1)
    zlim_coeff: float, opt
        corresponds to the zlim_coeff fraction of SN with z<zlim
    bands : `str`, opt
        bands to consider (default: grizy)
    gammaName: `str`, opt
        name of the gamma ref file to load (default: gamma_WFD.hdf5)
    dust : `bool`, opt
        Apply dust extinction to visit depth values (default False)
    hard_dust_cut : `float`, opt
      If set, cut any point on the sky that has an ebv extinction
      higher than the hard_dust_cut value.
      Default 0.25
    """

    def __init__(
        self,
        metric_name="SNNSNMetric",
        mjd_col="observationStartMJD",
        filter_col="filter",
        m5_col="fiveSigmaDepth",
        exptime_col="visitExposureTime",
        night_col="night",
        obsid_col="observationId",
        nexp_col="numExposures",
        vistime_col="visitTime",
        seeing_col="seeingFwhmEff",
        note_col="scheduler_note",
        season=[-1],
        coadd_night=True,
        zmin=0.1,
        zmax=0.5,
        z_step=0.03,
        daymax_step=3.0,
        verbose=False,
        ploteffi=False,
        n_bef=3,
        n_aft=8,
        snr_min=1.0,
        n_phase_min=1,
        n_phase_max=1,
        sigma_c=0.04,
        zlim_coeff=0.95,
        bands="grizy",
        add_dust=False,
        hard_dust_cut=0.25,
        gamma_name="gamma_WFD.hdf5",
        **kwargs,
    ):
        # n_bef / n_aft = 3/8 for WFD, 4/10 for DDF

        self.mjd_col = mjd_col
        self.m5_col = m5_col
        self.filter_col = filter_col
        self.exptime_col = exptime_col
        self.season_col = "season"
        self.night_col = night_col
        self.obsid_col = obsid_col
        self.nexp_col = nexp_col
        self.vistime_col = vistime_col
        self.seeing_col = seeing_col
        self.note_col = note_col

        self.ploteffi = ploteffi
        self.t0s = "all"
        self.zlim_coeff = zlim_coeff
        self.bands = bands
        self.coadd_night = coadd_night
        self.add_dust = add_dust
        self.hard_dust_cut = hard_dust_cut

        maps = ["DustMap"]
        dust_properties = DustValues()
        self.ax1 = dust_properties.ax1

        cols = [
            self.night_col,
            self.m5_col,
            self.filter_col,
            self.mjd_col,
            self.obsid_col,
            self.nexp_col,
            self.vistime_col,
            self.exptime_col,
            self.note_col,
        ]

        super(SNNSNMetric, self).__init__(
            col=cols,
            metric_dtype="object",
            metric_name=metric_name,
            maps=maps,
            **kwargs,
        )

        self.season = season

        # LC selection parameters
        self.n_bef = n_bef  # nb points before peak
        self.n_aft = n_aft  # nb points after peak
        self.snr_min = snr_min  # SNR cut for points before/after peak
        self.n_phase_min = n_phase_min  # nb of point with phase <=-5
        self.n_phase_max = n_phase_max  # nb of points with phase >=20
        self.sigma_c = sigma_c

        # loading reference LC files
        lc_reference = load_sne_cached(gamma_name)
        # loading reference LC files

        self.lc_fast = {}
        for key, vals in lc_reference.items():
            self.lc_fast[key] = LcfastNew(
                vals,
                key[0],
                key[1],
                self.mjd_col,
                self.filter_col,
                self.exptime_col,
                self.m5_col,
                self.season_col,
                self.nexp_col,
                self.seeing_col,
                self.snr_min,
                light_output=False,
            )
        # loading parameters
        self.zmin = zmin  # zmin for the study
        self.zmax = zmax  # zmax for the study
        self.zstep = z_step  # zstep
        # get redshift range for processing
        zrange = list(np.arange(self.zmin, self.zmax, self.zstep))
        if zrange[0] < 1.0e-6:
            zrange[0] = 0.01

        self.zrange = np.unique(zrange)

        self.daymax_step = daymax_step  # daymax step
        self.min_rf_phase = -20.0  # min ref phase for LC points selection
        self.max_rf_phase = 60.0  # max ref phase for LC points selection

        self.min_rf_phase_qual = -15.0  # min ref phase for bounds effects
        self.max_rf_phase_qual = 30.0  # max ref phase for bounds effects

        # snrate
        self.rate_sn = SnRate(
            h0=70.0,
            om0=0.3,
            min_rf_phase=self.min_rf_phase_qual,
            max_rf_phase=self.max_rf_phase_qual,
        )

        # verbose mode - useful for debug and code performance estimation
        self.verbose = verbose

        # supernovae parameters for fisher estimation
        self.params = ["x0", "x1", "daymax", "color"]

    def run(self, data_slice, slice_point):
        """
        Run method of the metric

        Parameters
        --------------
        data_slice : `np.ndarray`
            Observations to process (scheduler simulations)
        slice_point : `bool`, opt
            Information about the location on the sky from the slicer

        Returns
        -------
        metricVal : `np.ndarray`
            ['n_sn', 'zlim'] at this point on the sky
        """
        # Hard dust cut
        if self.hard_dust_cut is not None:
            ebvof_mw = slice_point["ebv"]
            if ebvof_mw > self.hard_dust_cut:
                return self.badval

        # get area on-sky in this slice_point
        if "nside" in slice_point:
            self.pix_area = hp.nside2pixarea(slice_point["nside"], degrees=True)
        else:
            self.pix_area = 9.6

        # If we want to apply dust extinction.
        if self.add_dust:
            new_m5 = data_slice[self.m5_col] * 0
            for filtername in np.unique(data_slice[self.filter_col]):
                in_filt = np.where(data_slice[self.filter_col] == filtername)[0]
                a_x = self.ax1[filtername] * slice_point["ebv"]
                new_m5[in_filt] = data_slice[self.m5_col][in_filt] - a_x
            data_slice[self.m5_col] = new_m5

        # select observations filter
        good_filters = np.isin(data_slice[self.filter_col], list(self.bands))
        data_slice = data_slice[good_filters]

        # coaddition per night and per band (if requested by the user)
        if self.coadd_night:
            data_slice = self.coadd(data_slice)

        # This seems unlikely, but is a possible bailout point
        if len(data_slice) <= self.n_aft + self.n_bef:
            return self.badval

        # get season information (seasons calculated by gaps,
        # not by place on sky)
        data_slice = self.getseason(data_slice, mjd_col=self.mjd_col)

        # get redshift values per season
        zseason = self.z_season(self.season, data_slice)
        zseason_allz = self.z_season_allz(zseason)

        # estimate redshift completeness
        metric_values = self.metric(data_slice, zseason_allz, x1=-2.0, color=0.2, zlim=-1, metric="zlim")

        if metric_values is None:
            return self.badval

        if np.max(metric_values["zcomp"]) < 0:
            return self.badval

        # get redshift values per season up to zcomp
        zseason = pd.DataFrame(metric_values[["season", "zcomp"]])
        zseason.loc[:, "zmin"] = 0.01
        zseason.loc[:, "zstep"] = self.zstep
        zseason = zseason.rename(columns={"zcomp": "zmax"})
        zseason["zmax"] += self.zstep
        zseason_allz = self.z_season_allz(zseason)

        # get the total number of well-sampled SN up to zcomp
        nsn_zcomp = self.metric(
            data_slice,
            zseason_allz,
            x1=0.0,
            color=0.0,
            zlim=metric_values[["season", "zcomp"]],
            metric="nsn",
        )

        # final results
        if nsn_zcomp is None:
            return self.badval
        metric_values = metric_values.merge(nsn_zcomp, left_on=["season"], right_on=["season"])

        if self.verbose:
            print("metric_values", metric_values[["season", "zcomp", "nsn"]])

        idx = metric_values["zcomp"] > 0.0
        selmet = metric_values[idx]

        if len(selmet) > 0:
            zcomp = selmet["zcomp"].median()
            n_sn = selmet["nsn"].sum()
            res = np.rec.fromrecords([(n_sn, zcomp)], names=["n_sn", "zlim"])
        else:
            res = self.badval

        if self.verbose:
            print("final result", res)

        return res

    def season_length(self, seasons, data_slice, zseason):
        """
        Method to estimate season lengths vs z

        Parameters
        -----------
        seasons : `list` [`int`]
            list of seasons to process
        data_slice : `np.ndarray`, (N,)`
            array of observations
        zseason : `pd.DataFrame`
            redshift infos per season

        Returns
        --------
        seasons : `list` [`int`]
            list of seasons to process
        dur_z : `pd.DataFrame`
            season lengths vs z
        """
        # if seasons = -1: process the seasons seen in data
        if seasons == [-1]:
            seasons = np.unique(data_slice[self.season_col])

        # season infos
        dfa = pd.DataFrame(np.copy(data_slice))
        dfa = pd.DataFrame(dfa[dfa["season"].isin(seasons)])

        season_info = self.get_season_info(dfa, zseason)

        if self.verbose:
            print("season_info", season_info)

        if season_info.empty:
            return [], pd.DataFrame()

        season_info["season_copy"] = season_info["season"].values.copy()
        dur_z = (
            season_info.groupby("season_copy", group_keys=False)
            .apply(lambda x: self.nsn_expected_z(x), include_groups=False)
            .reset_index(drop=True)
        )
        return season_info["season"].to_list(), dur_z

    def get_season_info(self, dfa, zseason, min_duration=60.0):
        """
        method to get season infos vs z

        Parameters
        --------------
        dfa : `pd.DataFrame`
            data to process
        zseason : `pd.DataFrame`
            redshift infos per season
        min_duration : `float`, opt
            min season length to be accepted (default: 60 days)

        Returns
        --------
        season_info : `pd.DataFrame` with season length infos

        """
        dfa["season_copy"] = dfa["season"].values.copy()
        season_info = (
            dfa.groupby("season")
            .apply(lambda x: self.season_info(x, min_duration=min_duration), include_groups=False)
            .reset_index()
        )

        # season_info.index = season_info.index.droplevel()
        season_info = season_info.drop(columns=["level_1"])

        season_info = season_info.merge(zseason, left_on=["season"], right_on=["season"])

        season_info["T0_min"] = season_info["MJD_min"] - (1.0 + season_info["z"]) * self.min_rf_phase_qual
        season_info["T0_max"] = season_info["MJD_max"] - (1.0 + season_info["z"]) * self.max_rf_phase_qual
        season_info["season_length"] = season_info["T0_max"] - season_info["T0_min"]

        idx = season_info["season_length"] >= min_duration

        return pd.DataFrame(season_info[idx])

    def step_lc(self, obs, gen_par, x1=-2.0, color=0.2):
        """
        Method to generate lc

        Parameters
        ---------------
        obs : array
            observations
        gen_par : array
            simulation parameters
        x1 : `float`, opt
            stretch value (default: -2.0)
        color : `float`, opt
            color value (default: 0.2)

        Returns
        ----------
        SN light curves (astropy table)

        """
        obs["season_copy"] = obs["season"].values.copy()
        lc = obs.groupby(["season_copy"]).apply(
            lambda x: self.gen_lc(x, gen_par, x1, color), include_groups=False
        )
        return lc

    def step_efficiencies(self, lc):
        """
        Method to estimate observing efficiencies

        Parameter
        -------------
        lc: `pd.DataFrame`
           light curves

        Returns
        -----------
        `pd.DataFrame` with efficiencies

        """
        cols = ["season", "z", "x1", "color", "sntype"]
        for col in cols:
            lc[col + "_copy"] = lc[col].values.copy()

        sn_effis = lc.groupby(cols).apply(lambda x: self.sn_effi(x), include_groups=False).reset_index()

        # estimate efficiencies
        sn_effis["season"] = sn_effis["season"].astype(int)
        sn_effis["effi"] = sn_effis["nsel"] / sn_effis["ntot"]
        sn_effis["effi_err"] = np.sqrt(sn_effis["nsel"] * (1.0 - sn_effis["effi"])) / sn_effis["ntot"]

        # prevent NaNs, set effi to 0 where there is 0 ntot
        zero = np.where(sn_effis["ntot"] == 0)
        sn_effis["effi"].values[zero] = 0
        sn_effis["effi_err"].values[zero] = 0

        if self.verbose:
            for season in sn_effis["season"].unique():
                idx = sn_effis["season"] == season
                print("effis", sn_effis[idx])

        if self.ploteffi:
            from sn_metrics.sn_plot_live import plotNSN_effi

            for season in sn_effis["season"].unique():
                idx = sn_effis["season"] == season
                print("effis", sn_effis[idx])
                plotNSN_effi(sn_effis[idx], "effi", "effi_err", "Observing Efficiencies", ls="-")

        return sn_effis

    def step_nsn(self, sn_effis, dur_z):
        """
        Method to estimate the number of supernovae from efficiencies

        Parameters
        ----------
        sn_effis : `pd.DataFrame`
          data with efficiencies of observation
        dur_z :  array
          array of season length

        Returns
        -------
        initial sn_effis appended with a set of infos (duration, nsn)

        """
        # add season length here
        sn_effis = sn_effis.merge(dur_z, left_on=["season", "z"], right_on=["season", "z"])

        # estimate the number of supernovae
        sn_effis["nsn"] = sn_effis["effi"] * sn_effis["nsn_expected"]

        return sn_effis

    def season_info(self, grp, min_duration):
        """
        Method to estimate seasonal info (cadence, season length, ...)

        Parameters
        --------------
        grp : `pd.DataFrame` group
        min_duration : `float`
          minimal duration for a season to be considered

        Returns
        ---------
        `pd.DataFrame` with the following cols:
        - Nvisits: number of visits for this group
        - N_xx:  number of visits in xx where xx is defined in self.bandstat

        """
        df = pd.DataFrame([len(grp)], columns=["Nvisits"])
        df["MJD_min"] = grp[self.mjd_col].min()
        df["MJD_max"] = grp[self.mjd_col].max()
        df["season_length"] = df["MJD_max"] - df["MJD_min"]
        df["cadence"] = 0.0

        if len(grp) > 5:
            # to = grp.groupby(['night'])[self.mjd_col].median().sort_values()
            # df['cadence'] = np.mean(to.diff())
            nights = np.sort(grp["night"].unique())
            diff = np.asarray(nights[1:] - nights[:-1])
            df["cadence"] = np.median(diff).item()

            # select seasons of at least 30 days
        idx = df["season_length"] >= min_duration

        return df[idx]

    def duration_z(self, grp, min_duration=60.0):
        """
        Method to estimate the season length vs redshift
        This is necessary to take into account boundary effects
        when estimating the number of SN that can be detected

        daymin, daymax = min and max MJD of a season
        T0_min(z) =  daymin-(1+z)*min_rf_phase_qual
        T0_max(z) =  daymax-(1+z)*max_rf_phase_qual
        season_length(z) = T0_max(z)-T0_min(z)

        Parameters
        --------------
        grp : `pd.DataFrame` group
          data to process: season infos
        min_duration : `float`, opt
          min season length for a season to be processed (deafult: 60 days)

        Returns
        ----------
        `pd.DataFrame` with season_length, z, T0_min and T0_max cols

        """
        ## IS THIS CALLED FROM ANYWHERE?
        daymin = grp["MJD_min"].values
        daymax = grp["MJD_max"].values
        dur_z = pd.DataFrame(self.zrange, columns=["z"])
        dur_z["T0_min"] = daymin - (1.0 + dur_z["z"]) * self.min_rf_phase_qual
        dur_z["T0_max"] = daymax - (1.0 + dur_z["z"]) * self.max_rf_phase_qual
        dur_z["season_length"] = dur_z["T0_max"] - dur_z["T0_min"]
        # dur_z['season_length_orig'] = daymax-daymin
        # dur_z['season_length_orig'] = [daymax-daymin]*len(self.zrange)
        nsn = self.nsn_from_rate(dur_z)
        if self.verbose:
            print("dur_z", dur_z)
            print("nsn expected", nsn)
        dur_z = dur_z.merge(nsn, left_on=["z"], right_on=["z"])

        idx = dur_z["season_length"] > min_duration
        sel = dur_z[idx]
        if len(sel) < 2:
            return pd.DataFrame()
        return dur_z

    def calc_daymax(self, grp, daymax_step):
        """
        Method to estimate T0 (daymax) values for simulation.

        Parameters
        --------------
        grp: group (`pd.DataFrame` sense)
            group of data to process with the following cols:
            t0_min: T0 min value (per season)
            t0_max: T0 max value (per season)
        daymax_step: `float`
            step for T0 simulation

        Returns
        ----------
        `pd.DataFrame` with daymax, min_rf_phase, max_rf_phase values

        """

        if self.t0s == "all":
            t0_max = grp["T0_max"].values
            t0_min = grp["T0_min"].values
            num = (t0_max - t0_min) / daymax_step
            if t0_max - t0_min > 10:
                df = pd.DataFrame(np.linspace(t0_min, t0_max, int(np.max(num))), columns=["daymax"])
            else:
                df = pd.DataFrame([-1], columns=["daymax"])
        else:
            df = pd.DataFrame([0.0], columns=["daymax"])

        df["minRFphase"] = self.min_rf_phase
        df["maxRFphase"] = self.max_rf_phase

        return df

    def gen_lc(self, grp, gen_par_orig, x1, color):
        """
        Method to generate light curves from observations

        Parameters
        ---------------
        grp : pd group
          observations to process
        gen_par_orig : `pd.DataFrame`
          simulation parameters
        x1 : `float`
          SN stretch
        color : `float`
          SN color

        Returns
        ----------
        light curves as `pd.DataFrame`

        """
        season = grp.name
        idx = gen_par_orig["season"] == season
        gen_par = gen_par_orig[idx].to_records(index=False)

        sntype = dict(zip([(-2.0, 0.2), (0.0, 0.0)], ["faint", "medium"]))
        res = pd.DataFrame()
        key = (np.round(x1, 1), np.round(color, 1))
        vals = self.lc_fast[key]

        gen_par_cp = gen_par.copy()
        if key == (-2.0, 0.2):
            idx = gen_par_cp["z"] < 0.9
            gen_par_cp = gen_par_cp[idx]
        lc = vals(grp.to_records(index=False), 0.0, gen_par_cp, bands="grizy")
        lc["x1"] = key[0]
        lc["color"] = key[1]
        lc["sntype"] = sntype[key]
        res = pd.concat((res, lc))

        return res

    def sn_effi(self, lc):
        """
        Method to transform LCs to supernovae

        Parameters
        ---------------
        lc : pd grp
          light curve

        Returns
        ----------
        `pd.DataFrame` of sn efficiencies vs z
        """

        if self.verbose:
            print("effi for", lc.name)

        lcarr = lc.to_records(index=False)

        idx = lcarr["snr_m5"] >= self.snr_min

        lcarr = np.copy(lcarr[idx])

        t0s = np.unique(lcarr["daymax"])
        t0s.sort()

        delta_t = lcarr["daymax"] - t0s[:, np.newaxis]

        flag = np.abs(delta_t) < 1.0e-5

        resdf = pd.DataFrame(t0s, columns=["daymax"])

        # get n_phase_min, n_phase_max
        for vv in [
            "n_phmin",
            "n_phmax",
            "F_x0x0",
            "F_x0x1",
            "F_x0daymax",
            "F_x0color",
            "F_x1x1",
            "F_x1daymax",
            "F_x1color",
            "F_daymaxdaymax",
            "F_daymaxcolor",
            "F_colorcolor",
        ]:
            resdf[vv] = self.get_sum(lcarr, vv, len(delta_t), flag)

        nights = np.tile(lcarr["night"], (len(delta_t), 1))
        phases = np.tile(lcarr["phase"], (len(delta_t), 1))

        flagph = phases >= 0.0
        resdf["nepochs_aft"] = self.get_epochs(nights, flag, flagph)
        flagph = phases <= 0.0
        resdf["nepochs_bef"] = self.get_epochs(nights, flag, flagph)

        # replace NaN by 0
        # solution from: https://stackoverflow.com/
        # questions/77900971/
        # pandas-futurewarning-downcasting-
        # object-dtype-arrays-on-fillna-ffill-bfill
        with pd.option_context("future.no_silent_downcasting", True):
            resdf = resdf.fillna(0).infer_objects(copy=False)

        # get selection efficiencies
        effis = self.efficiencies(resdf)

        return effis

    def get_sum(self, lcarr, varname, nvals, flag):
        """
        Method to get the sum of variables using broadcasting

        Parameters
        --------------
        lcarr : numpy array
          data to process
        varname : `str`
          col to process in lcarr
        nvals : `int`
          dimension for tiling
        flag : array(bool)
          flag to apply

        Returns
        ----------
        array: the sum of the corresponding variable

        """

        phmin = np.tile(lcarr[varname], (nvals, 1))
        n_phmin = np.ma.array(phmin, mask=~flag)
        n_phmin = n_phmin.sum(axis=1)

        return n_phmin

    def get_epochs(self, nights, flag, flagph):
        """
        Method to get the number of epochs

        Parameters
        ---------------
        nights : array
          night number array
        flag : array(bool)
          flag to apply
        flagph : array(bool)
          flag to apply

        Returns
        -----------
        array with the number of epochs

        """
        nights_cp = np.copy(nights)
        B = np.ma.array(nights_cp, mask=~(flag & flagph))
        B.sort(axis=1)
        C = np.diff(B, axis=1) > 0
        D = C.sum(axis=1) + 1
        return D

    def sigma_s_nparams(self, grp):
        """
        Method to estimate variances of SN parameters
        from inversion of the Fisher matrix

        Parameters
        ---------------
        grp: `pd.DataFrame` of flux derivatives wrt SN parameters
        Returns
        ----------
        Diagonal elements of the inverted matrix (as `pd.DataFrame`)
        """

        # params = ['x0', 'x1', 'daymax', 'color']
        parts = {}
        for ia, vala in enumerate(self.params):
            for jb, valb in enumerate(self.params):
                if jb >= ia:
                    parts[ia, jb] = grp["F_" + vala + valb]

        # print(parts)
        size = len(grp)
        npar = len(self.params)
        fisher__big = np.zeros((npar * size, npar * size))
        big__diag = np.zeros((npar * size, npar * size))
        big__diag = []

        for iv in range(size):
            for ia, vala in enumerate(self.params):
                for jb, valb in enumerate(self.params):
                    if jb >= ia:
                        fisher__big[ia + npar * iv][jb + npar * iv] = parts[ia, jb][iv]

        # pprint.pprint(fisher__big)

        fisher__big = fisher__big + np.triu(fisher__big, 1).T
        big__diag = np.diag(np.linalg.inv(fisher__big))

        res = pd.DataFrame()
        for ia, vala in enumerate(self.params):
            indices = range(ia, len(big__diag), npar)
            res["Cov_{}{}".format(vala, vala)] = np.take(big__diag, indices)

        return res

    def efficiencies(self, dfo):
        """ "
        Method to estimate selection efficiencies

        Parameters
        ---------------
        df: `pd.DataFrame`
          data to process

        """

        if self.verbose:
            print(
                "selection params",
                self.n_phase_min,
                self.n_phase_max,
                self.n_bef,
                self.n_aft,
            )
            print(dfo)
        df = pd.DataFrame(dfo)
        df["select"] = df["n_phmin"] >= self.n_phase_min
        df["select"] &= df["n_phmax"] >= self.n_phase_max
        df["select"] &= df["nepochs_bef"] >= self.n_bef
        df["select"] &= df["nepochs_aft"] >= self.n_aft
        df["select"] = df["select"].astype(int)
        df["Cov_colorcolor"] = 100.0

        idx = df["select"] == 1

        bad_sn = pd.DataFrame(df.loc[~idx])
        good_sn = pd.DataFrame()
        if len(df[idx]) > 0:
            good_sn = pd.DataFrame(df.loc[idx].reset_index())
            sigma__fisher = self.sigma_s_nparams(good_sn)
            good_sn["Cov_colorcolor"] = sigma__fisher["Cov_colorcolor"]

        all_sn = pd.concat((good_sn, bad_sn))
        all_sn["select"] &= all_sn["Cov_colorcolor"] <= self.sigma_c**2
        idx = all_sn["select"] == 1

        if self.verbose:
            if len(good_sn) > 0:
                print("good SN", len(good_sn), good_sn[["daymax", "Cov_colorcolor"]])
            else:
                print("no good SN")
        return pd.DataFrame({"ntot": [len(all_sn)], "nsel": [len(all_sn[idx])]})

    def metric(self, data_slice, zseason, x1=-2.0, color=0.2, zlim=-1, metric="zlim"):
        """
        Method to run the metric

        Parameters
        ---------------
        data_slice: array
          observations to use for processing
        zseason: array
          season infos (season length vs z)
        x1: float, opt
          SN stretch (default: -2.0)
        color: float, opt
          SN color (default: -0.2)
        zlim: float, opt
          redshift limit used to estimate NSN (default: -1)
        metric: str, opt
          metric to estimate [zlim or nsn] (default: zlim)


        """
        """
        snType = "medium"
        if np.abs(x1 + 2.0) <= 1.0e-5:
            snType = "faint"
        """

        # get the season durations
        seasons, dur_z = self.season_length(self.season, data_slice, zseason)
        if not seasons or dur_z.empty:
            return None

        # chek the redshift range per season
        dur_z = self.check_dur_z(dur_z)
        if dur_z.empty:
            return None

        # Stick a season column on in case it got killed by a groupby
        if "season" not in list(dur_z.columns):
            dur_z["season"] = dur_z["index"] * 0 + np.unique(seasons)

        # get simulation parameters
        if np.size(np.unique(seasons)) > 1:
            dur_z["z_copy"] = dur_z["z"].values.copy()
            dur_z["season_copy"] = dur_z["season"].values.copy()
            gen_par = (
                dur_z.groupby(["z", "season"])
                .apply(lambda x: self.calc_daymax(x, self.daymax_step), include_groups=False)
                .reset_index()
            )
        else:
            dur_z["z_copy"] = dur_z["z"].values.copy()
            gen_par = (
                dur_z.groupby(["z"])
                .apply(lambda x: self.calc_daymax(x, self.daymax_step), include_groups=False)
                .reset_index()
            )
            gen_par["season"] = gen_par["level_1"] * 0 + np.unique(seasons)

        if gen_par.empty:
            return None

        # select observations corresponding to seasons
        obs = pd.DataFrame(np.copy(data_slice))
        obs = obs[obs["season"].isin(seasons)]

        # metric values in a DataFrame
        metric_values = pd.DataFrame()

        # generate LC here
        lc = self.step_lc(obs, gen_par, x1=x1, color=color)

        if self.verbose:
            print("daymax values", lc["daymax"].unique(), len(lc["daymax"].unique()))
            print(
                lc[
                    [
                        "daymax",
                        "z",
                        "flux",
                        "fluxerr_photo",
                        "flux_e_sec",
                        "flux_5",
                        self.m5_col,
                    ]
                ]
            )

        if len(lc) == 0:
            return None

        # get observing efficiencies and build sn for metric
        lc.index = lc.index.droplevel()

        # estimate efficiencies
        sn_effis = self.step_efficiencies(lc)

        # estimate nsn
        sn = self.step_nsn(sn_effis, dur_z)

        # estimate redshift completeness
        if metric == "zlim":
            sn["season_copy"] = sn["season"].values.copy()
            metric_values = (
                sn.groupby(["season"]).apply(lambda x: self.zlim(x), include_groups=False).reset_index()
            )

        if metric == "nsn":
            sn = sn.merge(zlim, left_on=["season"], right_on=["season"])
            sn["season_copy"] = sn["season"].values.copy()
            metric_values = (
                sn.groupby(["season"]).apply(lambda x: self.nsn(x), include_groups=False).reset_index()
            )

        return metric_values

    def z_season(self, seasons, data_slice):
        """
        Fill the z values per season

        Parameters
        --------------
        seasons: list
          seasons to process
        data_slice: array
          data to process

        """
        # if seasons = -1: process the seasons seen in data
        if seasons == [-1]:
            seasons = np.unique(data_slice[self.season_col])

        # `pd.DataFrame` with zmin, zmax, zstep per season
        zseason = pd.DataFrame(seasons, columns=["season"])
        zseason["zmin"] = self.zmin
        zseason["zmax"] = self.zmax
        zseason["zstep"] = self.zstep

        return zseason

    def z_season_allz(self, zseason):

        zseason["season_copy"] = zseason["season"].values.copy()
        zseason_allz = (
            zseason.groupby(["season"])
            .apply(
                lambda x: pd.DataFrame(
                    {"z": list(np.arange(x["zmin"].mean(), x["zmax"].mean(), x["zstep"].mean()))}
                ),
                include_groups=False,
            )
            .reset_index()
        )

        return zseason_allz[["season", "z"]]

    def nsn_from_rate(self, grp):
        """
        Method to estimate the expected number of supernovae

        Parameters
        ---------------
        grp: `pd.DataFrame`
          data to process

        Returns
        -----------
        `pd.DataFrame` with z and nsn_expected as cols

        """
        durinterp_z = interp1d(
            grp["z"],
            grp["season_length"],
            kind="linear",
            bounds_error=False,
            fill_value=0,
        )
        zz, rate, err_rate, nsn, err_nsn = self.rate_sn(
            zmin=self.zmin,
            zmax=self.zmax,
            dz=self.zstep,
            duration_z=durinterp_z,
            # duration=self.duration_ref,
            survey_area=self.pix_area,
            account_for_edges=False,
        )

        nsn_expected = interp1d(zz, nsn, kind="linear", bounds_error=False, fill_value=0)
        nsn_res = nsn_expected(grp["z"])

        return pd.DataFrame({"nsn_expected": nsn_res, "z": grp["z"].to_list()})

    def coadd(self, obs):
        """
        Method to coadd data per band and per night

        Parameters
        ------------
        data : `pd.DataFrame`
            `pd.DataFrame` of observations

        Returns
        -------
        coadded data : `pd.DataFrame`

        """

        data = pd.DataFrame(np.copy(obs))
        keygroup = [self.filter_col, self.night_col]

        data.sort_values(by=keygroup, ascending=[True, True], inplace=True)

        # get the median single exptime
        exptime_single = data[self.exptime_col].median()
        coadd_df = (
            data.groupby(keygroup)
            .agg(
                {
                    self.nexp_col: ["sum"],
                    self.vistime_col: ["sum"],
                    self.exptime_col: ["sum"],
                    self.mjd_col: ["mean"],
                    self.m5_col: ["mean"],
                }
            )
            .reset_index()
        )

        coadd_df.columns = [
            self.filter_col,
            self.night_col,
            self.nexp_col,
            self.vistime_col,
            self.exptime_col,
            self.mjd_col,
            self.m5_col,
        ]

        coadd_df = coadd_df.sort_values(by=self.mjd_col)
        coadd_df[self.m5_col] += 1.25 * np.log10(coadd_df[self.exptime_col] / exptime_single)

        return coadd_df.to_records(index=False)

    def getseason(self, obs, season_gap=80.0, mjd_col="observationStartMJD"):
        """
        Method to estimate seasons

        Parameters
        ------------
        obs: `np.ndarray`
            array of observations
        season_gap: `float`, optional
            minimal gap required to define a season (default: 80 days)
        mjd_col: `str`, optional
            col name for MJD infos (default: observationStartMJD)

        Returns
        ---------
        obs : `np.ndarray`
            original numpy array with seasonnumber appended
        """

        # check whether season has already been estimated
        obs.sort(order=mjd_col)

        seasoncalc = np.ones(obs.size, dtype=int)

        if len(obs) > 1:
            diff = np.diff(obs[mjd_col])
            flag = np.where(diff > season_gap)[0]

            if len(flag) > 0:
                for i, indx in enumerate(flag):
                    seasoncalc[indx + 1 :] = i + 2

        obs = nlr.append_fields(obs, "season", seasoncalc)

        return obs

    def reducen_sn(self, metric_val):
        # At each slice_point, return the sum nSN value.

        return np.sum(metric_val["n_sn"])

    def reducezlim(self, metric_val):
        # At each slice_point, return the median zlim
        result = np.median(metric_val["zlim"])
        if result < 0:
            result = self.badval

        return result

    def nsn_expected_z(self, grp):
        """
        Method to estimate the expected nsn  vs redshift

        Parameters
        --------------
        grp: `pd.DataFrame` group
          data to process: season infos

        Returns
        ----------
        `pd.DataFrame` with season_length, z, nsn_expected cols

        """

        if len(grp) < 2:
            nsn = pd.DataFrame(grp["z"].to_list(), columns=["z"])
            nsn.loc[:, "nsn_expected"] = 0
        else:
            nsn = self.nsn_from_rate(grp)

        if self.verbose:
            print("dur_z", grp)
            print("nsn expected", nsn)

        dur_z = pd.DataFrame(grp)
        dur_z = dur_z.merge(nsn, left_on=["z"], right_on=["z"])

        # if "season" in dur_z.columns:
        #    dur_z = dur_z.drop(columns=["season"])

        return dur_z

    def zlim_or_nsn(self, effi, sntype="faint", zlim=-1.0):
        """
        Method to estimate the redshift limit or the number of sn

        Parameters
        ---------------
        effi : `pd.DataFrame`
            data to process
        sntype : `str`, opt
            type of SN to consider for estimation (default: faint)
        zlim : `float`, opt
            redshift limit

        Returns
        -----------
        if zlim<0: returns the redshift limit
        if zlim>0: returns the number of sn up to zlim


        """

        seleffi = effi[effi["sntype"] == sntype]
        seleffi = seleffi.reset_index(drop=True)
        seleffi = seleffi.sort_values(by=["z"])
        nsn_cum = np.cumsum(seleffi["nsn"].to_list())

        resa = -1.0
        if zlim < 0:
            df = pd.DataFrame(seleffi).reset_index(drop=True)
            df.loc[:, "nsn_cum"] = nsn_cum / nsn_cum[-1]
            index = df[df["nsn_cum"] < 1].index
            if len(index) == 0:
                return resa
            dfb = df[: index[-1] + 2]
            zlim = interp1d(
                dfb["nsn_cum"],
                dfb["z"],
                kind="linear",
                bounds_error=False,
                fill_value=0,
            )
            resa = zlim(self.zlim_coeff)
        else:
            effi = interp1d(
                seleffi["z"],
                seleffi["effi"],
                kind="linear",
                bounds_error=False,
                fill_value=0,
            )
            durinterp_z = interp1d(
                seleffi["z"],
                seleffi["season_length"],
                kind="linear",
                bounds_error=False,
                fill_value=0,
            )
            zmin, zstep = 0.1, 0.001
            resa = self.get_nsn(effi, durinterp_z, zmin, zlim, zstep)

        return np.round(resa, 6)

    def zlim(self, grp, sn_type="faint"):
        """
        Method to estimate the metric zcomp

        Parameters
        ---------------
        grp: pd group
        sn_type: str, opt
          type of SN to estimate zlim (default: faint)

        Returns
        ------------
        zcomp : `pd.DataFrame` with the metric as cols
        """
        zcomp = -1

        if grp["effi"].mean() > 0.02 and len(grp["effi"]) >= 2:
            zcomp = self.zlim_or_nsn(grp, sn_type, -1)

        return pd.DataFrame({"zcomp": [zcomp]})

    def nsn(self, grp, sn_type="medium"):
        """
        Method to estimate the metric nsn up to zlim

        Parameters
        ---------------
        grp: pd group
        sn_type: str, opt
          type of SN to estimate zlim (default: medium)

        Returns
        ------------
        nsn : `pd.DataFrame`
            Dataframe with the metric as cols
        """
        nsn = -1

        if grp["effi"].mean() > 0.02:
            nsn = self.zlim_or_nsn(grp, sn_type, grp["zcomp"].mean())

        return pd.DataFrame({"nsn": [nsn]})

    def get_nsn(self, effi, durinterp_z, zmin, zmax, zstep):
        """
        Method to estimate to total number of SN: NSN = Sum(effi(z)*rate(z))

        Parameters
        -----------
        effi : 1D interpolator
            efficiencies vs z
        durinterp_z : 1D interpolator
            duration vs z
        zmin : `float`
            redshift min
        zmax : `float`
            redshift max
        zstep : `float`
            redshift step

        Returns
        ----------
        tot_sn : `int`
            total number of SN up to zmax
        """

        zz, rate, err_rate, nsn, err_nsn = self.rate_sn(
            zmin=zmin,
            zmax=zmax + zstep,
            dz=zstep,
            duration_z=durinterp_z,
            # duration=180.,
            survey_area=self.pix_area,
            account_for_edges=False,
        )
        res = np.cumsum(effi(zz) * nsn)
        if np.size(res) > 0:
            return res[-1]
        else:
            return 0

    def check_dur_z(self, dur_z, nmin=2):
        """ "
        Method to remove seasons with a poor redshift range due
        to too low season length

        Parameters
        ----------------
        dur_z: `pd.DataFrame`
          data to process
        nmin: int, opt
          minimal number of redshift points per season (default: 2)

        Returns
        -----------
        dur_z_subset : `pd.DataFrame`
            dur_z but only with seasons having at least nmin points in redshift

        """

        dur_z["size"] = dur_z.groupby(["season"])["z"].transform("size")

        idx = dur_z["size"] >= nmin

        return pd.DataFrame(dur_z[idx])



__all__ = ("MapsRegistry", "BaseMap")

import inspect


class MapsRegistry(type):
    """
    Meta class for Maps, to build a registry of maps classes.
    """

    def __init__(cls, name, bases, dict):
        super(MapsRegistry, cls).__init__(name, bases, dict)
        if not hasattr(cls, "registry"):
            cls.registry = {}
        modname = inspect.getmodule(cls).__name__
        if modname.startswith("rubin_sim.maf.maps"):
            modname = ""
        else:
            if len(modname.split(".")) > 1:
                modname = ".".join(modname.split(".")[:-1]) + "."
            else:
                modname = modname + "."
        mapsname = modname + name
        if mapsname in cls.registry:
            raise Exception("Redefining maps %s! (there are >1 maps with the same name)" % (mapsname))
        if mapsname != "BaseMaps":
            cls.registry[mapsname] = cls

    def get_class(cls, mapsname):
        return cls.registry[mapsname]

    def help(cls, doc=False):
        for mapsname in sorted(cls.registry):
            if not doc:
                print(mapsname)
            if doc:
                print("---- ", mapsname, " ----")
                print(cls.registry[mapsname].__doc__)
                maps = cls.registry[mapsname]()
                print(" added to slice_point: ", ",".join(maps.keynames))


class BaseMap(metaclass=MapsRegistry):
    """Base for maps classes."""

    def __init__(self, **kwargs):
        self.keynames = ["newkey"]

    def __eq__(self, othermap):
        return self.keynames == othermap.keynames

    def __ne__(self, othermap):
        return self.keynames != othermap.keynames

    def __lt__(self, othermap):
        return self.keynames < othermap.keynames

    def __gt__(self, othermap):
        return self.keynames > othermap.keynames

    def __le__(self, othermap):
        return self.keynames <= othermap.keynames

    def __ge__(self, othermap):
        return self.keynames >= othermap.keynames

    def run(self, slice_points):
        """Given slice_points (dict containing metadata about each slice_point,
        including ra/dec), adds additional metadata at each slice_point
        and returns updated dict.
        """
        raise NotImplementedError("This must be defined in subclass")



import unittest

import numpy as np

import rubin_sim.maf.metrics as metrics
import rubin_sim.maf.stackers as stackers


class TestCalibrationMetrics(unittest.TestCase):
    def test_parallax_metric(self):
        """
        Test the parallax metric.
        """
        names = [
            "observationStartMJD",
            "finSeeing",
            "fiveSigmaDepth",
            "fieldRA",
            "fieldDec",
            "filter",
        ]
        types = [float, float, float, float, float, (np.str_, 1)]
        data = np.zeros(700, dtype=list(zip(names, types)))
        slice_point = {"sid": 0}
        data["observationStartMJD"] = np.arange(700) + 56762
        data["finSeeing"] = 0.7
        data["filter"][0:100] = str("r")
        data["filter"][100:200] = str("u")
        data["filter"][200:] = str("g")
        data["fiveSigmaDepth"] = 24.0
        stacker = stackers.ParallaxFactorStacker()
        data = stacker.run(data)
        norm_flags = [False, True]
        for flag in norm_flags:
            data["finSeeing"] = 0.7
            data["fiveSigmaDepth"] = 24.0
            baseline = metrics.ParallaxMetric(normalize=flag, seeing_col="finSeeing").run(data, slice_point)
            data["finSeeing"] = data["finSeeing"] + 0.3
            worse1 = metrics.ParallaxMetric(normalize=flag, seeing_col="finSeeing").run(data, slice_point)
            worse2 = metrics.ParallaxMetric(normalize=flag, rmag=22.0, seeing_col="finSeeing").run(
                data, slice_point
            )
            worse3 = metrics.ParallaxMetric(normalize=flag, rmag=22.0, seeing_col="finSeeing").run(
                data[0:300], slice_point
            )
            data["fiveSigmaDepth"] = data["fiveSigmaDepth"] - 1.0
            worse4 = metrics.ParallaxMetric(normalize=flag, rmag=22.0, seeing_col="finSeeing").run(
                data[0:300], slice_point
            )
            # Make sure the RMS increases as seeing increases,
            # the star gets fainter, the background gets brighter,
            # or the baseline decreases.
            if flag:
                pass
            else:
                assert worse1 > baseline
                assert worse2 > worse1
                assert worse3 > worse2
                assert worse4 > worse3

    def test_proper_motion_metric(self):
        """
        Test the ProperMotion metric.
        """
        names = [
            "observationStartMJD",
            "finSeeing",
            "fiveSigmaDepth",
            "fieldRA",
            "fieldDec",
            "filter",
        ]
        types = [float, float, float, float, float, (np.str_, 1)]
        data = np.zeros(700, dtype=list(zip(names, types)))
        slice_point = [0]
        stacker = stackers.ParallaxFactorStacker()
        norm_flags = [False, True]
        data["observationStartMJD"] = np.arange(700) + 56762
        data["finSeeing"] = 0.7
        data["filter"][0:100] = str("r")
        data["filter"][100:200] = str("u")
        data["filter"][200:] = str("g")
        data["fiveSigmaDepth"] = 24.0
        data = stacker.run(data)
        for flag in norm_flags:
            data["finSeeing"] = 0.7
            data["fiveSigmaDepth"] = 24
            baseline = metrics.ProperMotionMetric(normalize=flag, seeing_col="finSeeing").run(
                data, slice_point
            )
            data["finSeeing"] = data["finSeeing"] + 0.3
            worse1 = metrics.ProperMotionMetric(normalize=flag, seeing_col="finSeeing").run(data, slice_point)
            worse2 = metrics.ProperMotionMetric(normalize=flag, rmag=22.0, seeing_col="finSeeing").run(
                data, slice_point
            )
            worse3 = metrics.ProperMotionMetric(normalize=flag, rmag=22.0, seeing_col="finSeeing").run(
                data[0:300], slice_point
            )
            data["fiveSigmaDepth"] = data["fiveSigmaDepth"] - 1.0
            worse4 = metrics.ProperMotionMetric(normalize=flag, rmag=22.0, seeing_col="finSeeing").run(
                data[0:300], slice_point
            )
            # Make sure the RMS increases as seeing increases,
            # the star gets fainter, the background gets brighter,
            # or the baseline decreases.
            if flag:
                # When normalized, mag of star and m5 don't matter
                # (just scheduling).
                self.assertAlmostEqual(worse2, worse1)
                self.assertAlmostEqual(worse4, worse3)
                # But using fewer points should make proper motion worse.
                # survey assumed to have same seeing and limiting mags.
                assert worse3 < worse2
            else:
                assert worse1 > baseline
                assert worse2 > worse1
                assert worse3 > worse2
                assert worse4 > worse3

    def test_parallax_coverage_metric(self):
        """
        Test the parallax coverage
        """
        names = [
            "observationStartMJD",
            "finSeeing",
            "fiveSigmaDepth",
            "fieldRA",
            "fieldDec",
            "filter",
            "ra_pi_amp",
            "dec_pi_amp",
        ]
        types = [float, float, float, float, float, "<U1", float, float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        data["filter"] = "r"
        data["fiveSigmaDepth"] = 25.0
        data["ra_pi_amp"] = 1.0
        data["dec_pi_amp"] = 1.0

        # All the parallax amplitudes are the same, should return zero
        metric = metrics.ParallaxCoverageMetric(seeing_col="finSeeing")
        val = metric.run(data)
        assert val == 0

        # Half at (1,1), half at (0.5,0.5)
        data["ra_pi_amp"][0:50] = 1
        data["dec_pi_amp"][0:50] = 1
        data["ra_pi_amp"][50:] = -1
        data["dec_pi_amp"][50:] = -1
        val = metric.run(data)
        self.assertAlmostEqual(val, 2.0**0.5)

        data["ra_pi_amp"][0:50] = 0.5
        data["dec_pi_amp"][0:50] = 0.5
        data["ra_pi_amp"][50:] = -0.5
        data["dec_pi_amp"][50:] = -0.5
        val = metric.run(data)
        self.assertAlmostEqual(val, 0.5 * 2**0.5)

        data["ra_pi_amp"][0:50] = 1
        data["dec_pi_amp"][0:50] = 0
        data["ra_pi_amp"][50:] = -1
        data["dec_pi_amp"][50:] = 0
        val = metric.run(data)
        assert val == 1

    def test_parallax_dcr_degen_metric(self):
        """
        Test the parallax-DCR degeneracy metric
        """
        names = [
            "observationStartMJD",
            "finSeeing",
            "fiveSigmaDepth",
            "fieldRA",
            "fieldDec",
            "filter",
            "ra_pi_amp",
            "dec_pi_amp",
            "ra_dcr_amp",
            "dec_dcr_amp",
        ]
        types = [float, float, float, float, float, "<U1", float, float, float, float]
        data = np.zeros(100, dtype=list(zip(names, types)))
        data["filter"] = "r"
        data["fiveSigmaDepth"] = 25.0

        # Set so ra is perfecly correlated
        data["ra_pi_amp"] = 1.0
        data["dec_pi_amp"] = 0.01
        data["ra_dcr_amp"] = 0.2

        metric = metrics.ParallaxDcrDegenMetric(seeing_col="finSeeing")
        val = metric.run(data)
        np.testing.assert_almost_equal(np.abs(val), 1.0, decimal=2)

        # set so the offsets are always nearly perpendicular
        data["ra_pi_amp"] = 0.001
        data["dec_pi_amp"] = 1.0
        data["ra_dcr_amp"] = 0.2

        metric = metrics.ParallaxDcrDegenMetric(seeing_col="finSeeing")
        val = metric.run(data)
        np.testing.assert_almost_equal(val, 0.0, decimal=2)

        # Generate a random distribution that should have
        # little or no correlation
        rng = np.random.RandomState(42)

        data["ra_pi_amp"] = rng.rand(100) * 2 - 1.0
        data["dec_pi_amp"] = rng.rand(100) * 2 - 1.0
        data["ra_dcr_amp"] = rng.rand(100) * 2 - 1.0
        data["dec_dcr_amp"] = rng.rand(100) * 2 - 1.0

        val = metric.run(data)
        assert np.abs(val) < 0.2

    def test_radius_obs_metric(self):
        """
        Test the RadiusObsMetric
        """

        names = ["fieldRA", "fieldDec"]
        dt = ["float"] * 2
        data = np.zeros(3, dtype=list(zip(names, dt)))
        data["fieldDec"] = [-0.1, 0, 0.1]
        slice_point = {"ra": 0.0, "dec": 0.0}
        metric = metrics.RadiusObsMetric()
        result = metric.run(data, slice_point)
        for i, r in enumerate(result):
            np.testing.assert_almost_equal(r, abs(data["fieldDec"][i]))
        assert metric.reduce_mean(result) == np.mean(result)
        assert metric.reduce_rms(result) == np.std(result)
        np.testing.assert_almost_equal(
            metric.reduce_full_range(result),
            np.max(np.abs(data["fieldDec"])) - np.min(np.abs(data["fieldDec"])),
        )


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "create_constellation",
    "starlink_tles_v1",
    "starlink_tles_v2",
    "oneweb_tles",
    "Constellation",
    "sun_alt_limits",
)

import numpy as np
from astropy import constants as const
from astropy import units as u
from rubin_scheduler.utils import SURVEY_START_MJD, Site, gnomonic_project_toxy, point_to_line_distance
from shapely.geometry import LineString, Point
from skyfield.api import EarthSatellite, load, wgs84

MJDOFFSET = 2400000.5
mjd0 = SURVEY_START_MJD


def sun_alt_limits():
    """Return sun altitude limits (degrees) at which zero illuminated
    satellites above 20 degrees altitude result.

    Different constellations have different limits at which zero illumination
    above 20 degrees occurs.

    Returns
    -------
    sun_alt_limits : `dict` [`str`: `float`]
        Dict with satellite constellation name keys,
        altitude limits values (degrees).
    """
    # Estimated in sun_alts_limits.ipynb
    result = {"slv1": -36.0, "slv2": -36.0, "oneweb": -53.0}
    return result


def satellite_mean_motion(altitude, mu=const.GM_earth, r_earth=const.R_earth):
    """Calculate mean motion of satellites at a given altitude in Earth's
    gravitational field.
    See https://en.wikipedia.org/wiki/Mean_motion#Formulae

    Parameters
    ----------
    altitude : `float`
        Altitude of the satellite.
        Should be a float with astropy units attached.

    Returns
    -------
    mean_motion : `float`
    """
    no = np.sqrt(4.0 * np.pi**2 * (altitude + r_earth) ** 3 / mu).to(u.day)
    return 1 / no


def tle_from_orbital_parameters(sat_name, sat_nr, epoch, inclination, raan, mean_anomaly, mean_motion):
    """Generate TLE strings from orbital parameters.

    Parameters
    ----------
    sat_name : `str`
        Satellite name
    sat_nr : `float`
        Satellite nr
    epoch : `float`
        Epoch
    inclination : `float`
        Inclination, probably with astropy units attached
    raan : `float`
        RA of the ascending node. Assumes astropy units.
    mean_anomaly : `float`
        Mean anomaly.
    mean_motion : `float`
       Mean motion.

    Returns
    -------
    tle : `str`

    Notes
    -----
    epoch has the format: first two digits are the year,
    next three digits are the day from beginning of year,
    then fraction of a day is given, e.g.
    20180.25 would be 2020, day 180, 6 hours (UT?)
    """

    # Note: RAAN = right ascention (or longitude) of ascending node
    # I suspect this is filling in 0 eccentricity everywhere.

    def checksum(line):
        s = 0
        for c in line[:-1]:
            if c.isdigit():
                s += int(c)
            if c == "-":
                s += 1
        return "{:s}{:1d}".format(line[:-1], s % 10)

    tle0 = sat_name
    tle1 = checksum(
        "1 {:05d}U 20001A   {:14.8f}  .00000000  00000-0  50000-4 " "0    0X".format(sat_nr, epoch)
    )
    tle2 = checksum(
        "2 {:05d} {:8.4f} {:8.4f} 0001000   0.0000 {:8.4f} "
        "{:11.8f}    0X".format(
            sat_nr,
            inclination.to_value(u.deg),
            raan.to_value(u.deg),
            mean_anomaly.to_value(u.deg),
            mean_motion.to_value(1 / u.day),
        )
    )

    return "\n".join([tle0, tle1, tle2])


def create_constellation(
    altitudes,
    inclinations,
    nplanes,
    sats_per_plane,
    epoch=23274.0,
    name="Test",
    seed=42,
):
    """Create a set of orbital elements for a satellite constellation then
    convert them to TLEs.

    Parameters
    ----------
    altitudes : `np.ndarray`, (N,)
        Altitudes (degrees).
    inclinations : `np.ndarray`, (N,)
        Inclinations (degrees).
    nplanes : `np.ndarray`, (N,)
        Number of satellite planes.
    sats_per_plane : `np.ndarray`, (N,)
        Number of satellites per orbital plane.
    epoch : `float`
        Epoch.
    name : `str`
        Satellite name.
    seed : `float`
        Random number seed.

    Returns
    -------
    my_sat_tles : `list` of `str`
    """

    rng = np.random.default_rng(seed)

    my_sat_tles = []
    sat_nr = 8000
    for alt, inc, n, s in zip(altitudes, inclinations, nplanes, sats_per_plane):
        if s == 1:
            # random placement for lower orbits
            mas = rng.uniform(0, 360, n) * u.deg
            raans = rng.uniform(0, 360, n) * u.deg
        else:
            mas = np.linspace(0.0, 360.0, s, endpoint=False) * u.deg
            mas += rng.uniform(0, 360, 1) * u.deg
            raans = np.linspace(0.0, 360.0, n, endpoint=False) * u.deg
            mas, raans = np.meshgrid(mas, raans)
            mas, raans = mas.flatten(), raans.flatten()

        mm = satellite_mean_motion(alt)
        for ma, raan in zip(mas, raans):
            my_sat_tles.append(
                tle_from_orbital_parameters(name + " {:d}".format(sat_nr), sat_nr, epoch, inc, raan, ma, mm)
            )
            sat_nr += 1

    return my_sat_tles


def starlink_tles_v1():
    """Create a list of satellite TLE's, appropriate for
    Starlink v1 (as of July 2022). Should create 4,408 orbits

    Returns
    -------
    my_sat_tles : `list` [`str`]
    """
    altitudes = np.array([550, 540, 570, 560, 560]) * u.km
    inclinations = np.array([53, 53.2, 70, 97.6, 97.6]) * u.deg
    nplanes = np.array([72, 72, 36, 6, 4])
    sats_per_plane = np.array([22, 22, 20, 58, 43])

    my_sat_tles = create_constellation(altitudes, inclinations, nplanes, sats_per_plane, name="starV1")

    return my_sat_tles


def starlink_tles_v2():
    """Create a list of satellite TLE's appropriate for
    Starlink v2 (as of July 2022). Should create 29,988 orbits

    Returns
    -------
    my_sat_tles : `list` [`str`]
    """
    altitudes = np.array([340, 345, 350, 360, 525, 530, 535, 604, 614]) * u.km
    inclinations = np.array([53, 46, 38, 96.9, 53, 43, 33, 148, 115.7]) * u.deg
    nplanes = np.array([48, 48, 48, 30, 28, 28, 28, 12, 18])
    sats_per_plane = np.array([110, 110, 110, 120, 120, 120, 120, 12, 18])

    my_sat_tles = create_constellation(altitudes, inclinations, nplanes, sats_per_plane, name="starV2")

    return my_sat_tles


def oneweb_tles():
    """Create a list of satellite TLE's appropriate
    for OneWeb plans (as of July 2022). Should create 6,372 orbits

    Returns
    -------
    my_sat_tles : `list` [`str`]
    """
    altitudes = np.array([1200, 1200, 1200]) * u.km
    inclinations = np.array([87.9, 40, 55]) * u.deg
    nplanes = np.array([49, 72, 72])
    sats_per_plane = np.array([36, 32, 32])

    my_sat_tles = create_constellation(altitudes, inclinations, nplanes, sats_per_plane, name="oneWe")

    return my_sat_tles


class Constellation:
    """Holds the constellation TLEs and calculates their appearance
    in a series of observations.

    Parameters
    ----------
    sat_tle_list : `list` [`str`]
        A list of satellite TLEs to be used
    alt_limit : `float`
        Altitude limit below which satellites can be ignored (degrees)
    fov : `float`
        The field of view diameter (degrees)
    """

    def __init__(self, sat_tle_list, alt_limit=20.0, fov=3.5):
        self.alt_limit_rad = np.radians(alt_limit)
        self.fov_radius_rad = np.radians(fov / 2.0)

        # Load ephemeris for sun position
        self.eph = load("de421.bsp")

        self.sat_list = []
        self.ts = load.timescale()
        for tle in sat_tle_list:
            name, line1, line2 = tle.split("\n")
            self.sat_list.append(EarthSatellite(line1, line2, name, self.ts))

        self._make_location()

    def _make_location(self):
        telescope = Site(name="LSST")

        self.observatory_site = wgs84.latlon(telescope.latitude, telescope.longitude, telescope.height)

    def update_mjd(self, mjd):
        """Calculate and record the alt/az position and illumination status
        for all the satellites at a given time.

        Parameters
        ----------
        mjd : `float`
            New MJD.
        """
        jd = mjd + MJDOFFSET
        t = self.ts.ut1_jd(jd)

        self.altitudes_rad = []
        self.azimuth_rad = []
        self.illum = []
        for sat in self.sat_list:
            current_sat = sat.at(t)
            illum = current_sat.is_sunlit(self.eph)
            self.illum.append(illum.copy())
            if illum:
                topo = current_sat - self.observatory_site.at(t)
                # this returns an Angle object
                alt, az, dist = topo.altaz()
                self.altitudes_rad.append(alt.radians + 0)
                self.azimuth_rad.append(az.radians + 0)
            else:
                self.altitudes_rad.append(np.nan)
                self.azimuth_rad.append(np.nan)

        self.altitudes_rad = np.array(self.altitudes_rad)
        self.azimuth_rad = np.array(self.azimuth_rad)
        self.illum = np.array(self.illum)
        # Keep track of the ones that are up and illuminated
        self.visible = np.where((self.altitudes_rad >= self.alt_limit_rad) & (self.illum == True))[0]

    def paths_array(self, mjds):
        """Calculate and return the RA/Dec/Alt and illumination status
        for all the satellites at an array of times.

        Parameters
        ----------
        mjds : `np.ndarray`, (N,)
            Modified Julian Dates.

        Returns
        -------
        ras : `np.ndarray`, (N,)
            RAs at each MJD
        decs : `np.ndarray`, (N,)
            Decs at each MJD
        alts : `np.ndarray`, (N,)
            Altitudes at each MJD
        illums : `np.ndarray`, (N,)
            Array of bools for if satellite is illuminated
        """

        jd = mjds + MJDOFFSET
        t = self.ts.ut1_jd(jd)

        ras = []
        decs = []
        alts = []
        illums = []
        for sat in self.sat_list:
            current_sat = sat.at(t)
            illum = current_sat.is_sunlit(self.eph)
            illums.append(illum.copy())
            topo = current_sat - self.observatory_site.at(t)
            ra, dec, distance = topo.radec()
            alt, az, dist = topo.altaz()
            ras.append(ra.radians)
            decs.append(dec.radians)
            alts.append(alt.radians)
        return np.vstack(ras), np.vstack(decs), np.vstack(alts), np.vstack(illums)

    def check_pointings(
        self,
        pointing_ras,
        pointing_decs,
        mjds,
        visit_time,
        fov_radius=1.75,
        test_radius=10.0,
        dt=2.0,
    ):
        """Calculate streak length and number of streaks in a set of visits.

        Parameters
        ----------
        pointing_ras : `np.ndarray`, (N,)
            The RA for each pointing (degrees).
        pointing_decs : `np.ndarray`, (N,)
            The dec for each pointing (degrees).
        mjds : `np.ndarray`, (N,)
            The MJD for the (start) of each pointing (days).
        visit_time : `np.ndarray`, (N,)
            The start to end time for a visit (seconds).
        fov_radius : `float`
            The radius of the science field of view (degrees)
        test_radius : `float`
            The radius to use to see if a streak gets close (degrees).
            Should be large, because satellites can be moving at ~1 deg/s.
        dt : `float`
            The timestep to use for high resolution checking
            if a satellite crossed (seconds).

        Returns
        -------
        streak length : `float`
            The total length of satellite streaks in the FoV (degrees)
        n_streak : `int`
            The number of streaks that were in the FoV.
        """
        test_radius_rad = np.radians(test_radius)
        dt = dt / 3600 / 24  # to days
        visit_time = visit_time / 3600.0 / 24.0

        # Arrays to hold results
        lengths_rad = np.zeros(pointing_ras.size, dtype=float)
        n_streaks = np.zeros(pointing_ras.size, dtype=int)

        input_id_indx_oned = np.arange(pointing_ras.size, dtype=int)

        # Convert everything to radians for internal computations
        pointing_ras_rad = np.radians(pointing_ras)
        pointing_decs_rad = np.radians(pointing_decs)
        fov_radius_rad = np.radians(fov_radius)

        # Note self.paths_array should return an array that is
        # N_sats x N_mjds in shape
        # And all angles in radians.
        sat_ra_1, sat_dec_1, sat_alt_1, sat_illum_1 = self.paths_array(mjds)
        mjd_end = mjds + visit_time
        sat_ra_2, sat_dec_2, sat_alt_2, sat_illum_2 = self.paths_array(mjd_end)

        # broadcast the pointings to be the same shape as the satellite arrays.
        pointing_ras_rad = np.broadcast_to(pointing_ras_rad, sat_ra_1.shape)
        pointing_decs_rad = np.broadcast_to(pointing_decs_rad, sat_ra_1.shape)
        input_id_indx = np.broadcast_to(input_id_indx_oned, sat_ra_1.shape)

        # Which satellites are above the altitude limit and illuminated
        # np.where confuses me when used on a 2d array.
        above_illum_indx = np.where(
            ((sat_alt_1 > self.alt_limit_rad) | (sat_alt_2 > self.alt_limit_rad))
            & ((sat_illum_1 == True) | (sat_illum_2 == True))
        )

        # point_to_line_distance can take arrays,
        # but they all need to be the same shape,
        # thus why we broadcast pointing ra and dec above.
        distances = point_to_line_distance(
            sat_ra_1[above_illum_indx],
            sat_dec_1[above_illum_indx],
            sat_ra_2[above_illum_indx],
            sat_dec_2[above_illum_indx],
            pointing_ras_rad[above_illum_indx],
            pointing_decs_rad[above_illum_indx],
        )

        close = np.where(distances < test_radius_rad)[0]

        # Numpy broadcasting is such a dark art
        sat_indx = np.arange(len(self.sat_list), dtype=int)[np.newaxis]
        sat_indx = np.broadcast_to(sat_indx.T, sat_ra_1.shape)

        mjd_broad = np.broadcast_to(mjds, sat_ra_1.shape)[above_illum_indx][close]
        visit_broad = np.broadcast_to(visit_time, sat_ra_1.shape)[above_illum_indx][close]

        # ok, this is pretty ugly, but should get the job done
        # Loop over all the potential collisions we have found
        for p_ra, p_dec, ob_indx, mjd, vt, sat_in in zip(
            pointing_ras_rad[above_illum_indx][close],
            pointing_decs_rad[above_illum_indx][close],
            input_id_indx[above_illum_indx][close],
            mjd_broad,
            visit_broad,
            sat_indx[above_illum_indx][close],
        ):
            mjd = np.linspace(mjd, mjd + vt, num=np.round(vt / dt).astype(int))
            jd = mjd + MJDOFFSET
            t = self.ts.ut1_jd(jd)
            sat = self.sat_list[sat_in]
            current_sat = sat.at(t)
            topo = current_sat - self.observatory_site.at(t)
            sat_ra, sat_dec, _distance = topo.radec()

            length = _streak_length(sat_ra.radians, sat_dec.radians, p_ra, p_dec, fov_radius_rad)
            if length > 0:
                lengths_rad[ob_indx] += length
                n_streaks[ob_indx] += 1
        return np.degrees(lengths_rad), n_streaks


def _streak_length(sat_ras, sat_decs, pointing_ra, pointing_dec, radius):
    """Calculate streak lengths for satellites in a given (circular) pointing.

    Parameters
    ----------
    sat_ras : `np.ndarray`, (N,)
        RA for each satellite (radians).
    sat_decs : `np.ndarray`, (N,)
        Decs for the satelltes (radians).
    pointing_ra : `float`
        RA of the pointing (radians).
    pointing_dec : `float`
        Dec of the pointing (radians).
    radius : `float`
        Radius of the field of view (radians).

    Returns
    -------
    length : `float`
        The length of the streak in radians
    """
    # Hopefully this broadcasts properly
    x, y = gnomonic_project_toxy(sat_ras, sat_decs, pointing_ra, pointing_dec)
    ls = LineString(zip(x, y))
    p = Point(0, 0)
    circle_buffer = p.buffer(radius)
    length = circle_buffer.intersection(ls).length
    return length



import os
import unittest

import numpy as np
import pandas as pd
from astropy.time import Time
from rubin_scheduler.data import get_data_dir

from rubin_sim.moving_objects import Orbits, PyOrbEphemerides


@unittest.skip("Temporary skip until ephemerides replaced")
class TestPyOrbEphemerides(unittest.TestCase):
    def setUp(self):
        self.testdir = os.path.join(get_data_dir(), "tests", "orbits_testdata")
        self.orbits = Orbits()
        self.orbits.read_orbits(os.path.join(self.testdir, "test_orbitsQ.des"))
        self.orbits_kep = Orbits()
        self.orbits_kep.read_orbits(os.path.join(self.testdir, "test_orbitsA.des"))
        self.ephems = PyOrbEphemerides()
        self.ephems.set_orbits(self.orbits)
        self.len_ephems_basic = 11
        self.len_ephems_full = 34

    def tear_down(self):
        del self.orbits
        del self.orbits_kep
        del self.ephems

    def test_which_pyoorb(self):
        import pyoorb

        print(pyoorb.__file__)

    def test_set_orbits(self):
        # Test that we can set orbits.
        self.ephems.set_orbits(self.orbits)
        # Test that setting with an empty orbit object fails.
        # (Avoids hard-to-interpret errors from pyoorb).
        with self.assertRaises(ValueError):
            empty_orb = Orbits()
            empty = pd.DataFrame([], columns=self.orbits.data_cols["KEP"])
            empty_orb.set_orbits(empty)
            self.ephems.set_orbits(empty_orb)

    def test_convert_to_oorb_array(self):
        # Check that orbital elements are converted.
        self.ephems._convert_to_oorb_elem(self.orbits.orbits, self.orbits.orb_format)
        self.assertEqual(len(self.ephems.oorb_elem), len(self.orbits))
        self.assertEqual(self.ephems.oorb_elem[0][7], 2)
        self.assertEqual(self.ephems.oorb_elem[0][9], 3)
        self.assertEqual(self.ephems.oorb_elem[0][1], self.orbits.orbits["q"][0])
        # Test that we can convert KEP orbital elements too.
        self.ephems._convert_to_oorb_elem(self.orbits_kep.orbits, self.orbits_kep.orb_format)
        self.assertEqual(len(self.ephems.oorb_elem), len(self.orbits_kep))
        self.assertEqual(self.ephems.oorb_elem[0][7], 3)
        self.assertEqual(self.ephems.oorb_elem[0][1], self.orbits_kep.orbits["a"][0])

    def test_convert_from_oorb_array(self):
        # Check that we can convert orbital elements TO oorb format and back
        # without losing info
        # (except ObjId -- we will lose that unless we use updateOrbits.)
        self.ephems._convert_to_oorb_elem(self.orbits.orbits, self.orbits.orb_format)
        new_orbits = Orbits()
        new_orbits.set_orbits(self.orbits.orbits)
        new_orbits.update_orbits(self.ephems.convert_from_oorb_elem())
        self.assertEqual(new_orbits, self.orbits)

    def test_convert_times(self):
        times = np.arange(49353, 49353 + 10, 0.5)
        eph_times = self.ephems._convert_times(times, "UTC")
        # Check that shape of eph_times is correct. (times x 2)
        self.assertEqual(eph_times.shape[0], len(times))
        self.assertEqual(eph_times.shape[1], 2)
        # Check that 'timescale' for eph_times is correct.
        self.assertEqual(eph_times[0][1], 1)
        eph_times = self.ephems._convert_times(times, "TAI")
        self.assertEqual(eph_times[0][1], 4)

    def test_oorb_ephemeris(self):
        self.ephems.set_orbits(self.orbits)
        times = np.arange(49353, 49353 + 3, 0.25)
        eph_times = self.ephems._convert_times(times)
        # Basic ephemerides.
        oorb_ephs = self.ephems._generate_oorb_ephs_basic(eph_times, obscode=807, eph_mode="N")
        # Check that it returned the right sort of array.
        self.assertEqual(
            oorb_ephs.shape,
            (len(self.ephems.oorb_elem), len(times), self.len_ephems_basic),
        )
        # Full ephemerides
        oorb_ephs = self.ephems._generate_oorb_ephs_full(eph_times, obscode=807, eph_mode="N")
        # Check that it returned the right sort of array.
        self.assertEqual(
            oorb_ephs.shape,
            (len(self.ephems.oorb_elem), len(times), self.len_ephems_full),
        )

    def test_ephemeris(self):
        # Calculate and convert ephemerides.
        self.ephems.set_orbits(self.orbits)
        times = np.arange(49353, 49353 + 2, 0.3)
        eph_times = self.ephems._convert_times(times)
        oorb_ephs = self.ephems._generate_oorb_ephs_basic(eph_times, obscode=807)
        # Group by object, and check grouping.
        ephs = self.ephems._convert_oorb_ephs_basic(oorb_ephs, by_object=True)
        self.assertEqual(len(ephs), len(self.orbits))
        # Group by time, and check grouping.
        oorb_ephs = self.ephems._generate_oorb_ephs_basic(eph_times, obscode=807)
        ephs = self.ephems._convert_oorb_ephs_basic(oorb_ephs, by_object=False)
        self.assertEqual(len(ephs), len(times))
        # And test all-wrapped-up method:
        ephs_all = self.ephems.generate_ephemerides(
            times,
            obscode=807,
            eph_mode="N",
            eph_type="basic",
            time_scale="UTC",
            by_object=False,
        )
        # See https://rubinobs.atlassian.net/browse/SP-1633
        # This needs to be fixed, but on a separate ticket
        # for key in ephs_all.dtype.names:
        #    np.testing.assert_almost_equal(ephs_all[key], ephs[key])

        # Reset ephems to use KEP Orbits, and calculate new ephemerides.
        self.ephems.set_orbits(self.orbits_kep)
        oorb_ephs = self.ephems._generate_oorb_ephs_basic(eph_times, obscode=807, eph_mode="N")
        ephs_kep = self.ephems._convert_oorb_ephs_basic(oorb_ephs, by_object=True)
        self.assertEqual(len(ephs_kep), len(self.orbits_kep))
        oorb_ephs = self.ephems._generate_oorb_ephs_basic(eph_times, obscode=807, eph_mode="N")
        ephs_kep = self.ephems._convert_oorb_ephs_basic(oorb_ephs, by_object=False)
        self.assertEqual(len(ephs_kep), len(times))
        # And test all-wrapped-up method:
        ephs_all_kep = self.ephems.generate_ephemerides(
            times,
            obscode=807,
            eph_mode="N",
            eph_type="basic",
            time_scale="UTC",
            by_object=False,
        )
        # Also https://rubinobs.atlassian.net/browse/SP-1633
        # for key in ephs_all_kep.dtype.names:
        #    np.testing.assert_almost_equal(ephs_all_kep[key], ephs_kep[key])

        # Check that ephemerides calculated from the different (COM/KEP)
        # orbits are almost equal
        for column in ephs.dtype.names:
            np.testing.assert_allclose(ephs[column], ephs_kep[column], rtol=1e-5, atol=1e-4)
        # Check that the wrapped method using KEP elements and the wrapped
        # method using COM elements match.
        for column in ephs_all.dtype.names:
            np.testing.assert_allclose(ephs_all_kep[column], ephs_all[column], rtol=1e-5, atol=1e-4)


@unittest.skip("Temporary skip until ephemerides replaced")
class TestJPLValues(unittest.TestCase):
    """Test the oorb generated RA/Dec values against
    JPL generated RA/Dec values."""

    def setUp(self):
        # Read orbits.
        self.orbits = Orbits()
        self.jpl_dir = os.path.join(get_data_dir(), "tests", "jpl_testdata")
        self.orbits.read_orbits(os.path.join(self.jpl_dir, "S0_n747.des"), skiprows=1)
        # Read JPL ephems.
        self.jpl = pd.read_csv(os.path.join(self.jpl_dir, "807_n747.txt"), sep=r"\s+")
        # Temp key fix
        self.jpl["obj_id"] = self.jpl["objId"]
        # Add times in TAI and UTC, because.
        t = Time(self.jpl["epoch_mjd"], format="mjd", scale="utc")
        self.jpl["mjdTAI"] = t.tai.mjd
        self.jpl["mjdUTC"] = t.utc.mjd

    def tear_down(self):
        del self.orbits
        del self.jpl

    def test_ra_dec(self):
        # We won't compare Vmag, because this also needs information
        # on trailing losses.
        times = self.jpl["mjdUTC"].unique()
        delta_ra = np.zeros(len(times), float)
        delta_dec = np.zeros(len(times), float)
        for i, t in enumerate(times):
            # Find the JPL obj_ids visible at this time.
            j = self.jpl.query("mjdUTC == @t").sort_values("obj_id")
            # Set the ephems, using the objects seen at this time.
            suborbits = self.orbits.orbits.query("obj_id in @j.obj_id").sort_values("obj_id")
            sub_orbits = Orbits()
            sub_orbits.set_orbits(suborbits)
            ephems = PyOrbEphemerides()
            ephems.set_orbits(sub_orbits)
            ephs = ephems.generate_ephemerides(
                [t],
                time_scale="UTC",
                obscode=807,
                eph_mode="N",
                eph_type="Basic",
                by_object=False,
            )
            delta_ra[i] = np.abs(ephs["ra"] - j["ra_deg"].values).max()
            delta_dec[i] = np.abs(ephs["dec"] - j["dec_deg"].values).max()
        # Convert to mas
        delta_ra *= 3600.0 * 1000.0
        delta_dec *= 3600.0 * 1000.0
        # Much of the time we're closer than 1mas,
        # but there are a few which hit higher values.
        print("max JPL errors", np.max(delta_ra), np.max(delta_dec))
        print("std JPL errors", np.std(delta_ra), np.std(delta_dec))
        self.assertLess(np.max(delta_ra), 25)
        self.assertLess(np.max(delta_dec), 25)
        self.assertLess(np.std(delta_ra), 3)
        self.assertLess(np.std(delta_dec), 3)


if __name__ == "__main__":
    unittest.main()



__all__ = (
    "fwhm_eff2_fwhm_geom",
    "fwhm_geom2_fwhm_eff",
    "calc_neff",
    "calc_instr_noise_sq",
    "calc_total_non_source_noise_sq",
    "calc_snr_sed",
    "calc_m5",
    "calc_sky_counts_per_pixel_for_m5",
    "calc_gamma",
    "calc_snr_m5",
    "calc_astrometric_error",
    "mag_error_from_snr",
    "calc_mag_error_m5",
    "calc_mag_error_sed",
    "scale_sky_m5",
)

import numpy

from .sed import Sed


def fwhm_eff2_fwhm_geom(fwhm_eff):
    """Convert fwhm_eff to fwhm_geom.

    This conversion was calculated by Bo Xin and Zeljko Ivezic
    (and will be in an update on the LSE-40 and overview papers).

    Parameters
    ----------
    fwhm_eff: `float`
        the single-gaussian equivalent FWHM value, appropriate for calc_neff,
        in arcseconds

    Returns
    -------
    fwhm_geom : `float`
        FWHM geom, the geometric FWHM value as measured from a typical
        PSF profile in arcseconds.
    """
    fwhm_geom = 0.822 * fwhm_eff + 0.052
    return fwhm_geom


def fwhm_geom2_fwhm_eff(fwhm_geom):
    """Convert fwhm_geom to fwhm_eff.

    This conversion was calculated by Bo Xin and Zeljko Ivezic
    (and will be in an update on the LSE-40 and overview papers).

    Parameters
    ----------
    fwhm_geom: `float`
        The geometric FWHM value, as measured from a typical PSF profile,
         in arcseconds.

    Returns
    -------
    fwhm_eff: `float`
        FWHM effective, the single-gaussian equivalent FWHM value,
        appropriate for calc_neff, in arcseconds.
    """
    fwhm_eff = (fwhm_geom - 0.052) / 0.822
    return fwhm_eff


def calc_neff(fwhm_eff, platescale):
    """Calculate the effective number of pixels in a single gaussian PSF.

    This equation comes from LSE-40, equation 27.
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40

    Parameters
    ----------
    fwhm_eff: `float`
        The width of a single-gaussian that produces correct
        Neff for typical PSF profile.
    platescale: `float`
        The platescale in arcseconds per pixel (0.2 for LSST)

    Returns
    -------
    nEff : `float`
        The effective number of pixels contained in the PSF

    Notes
    -----
    The fwhm_eff is a way to represent the equivalent seeing value, if the
    atmosphere could be simply represented as a single gaussian (instead of
    a more complicated von Karman profile for the atmosphere, convolved
    properly with the telescope hardware additional blurring of 0.4").
    A translation from the geometric FWHM to the fwhm_eff is provided
    in fwhm_geom2_fwhm_eff.
    """
    return 2.266 * (fwhm_eff / platescale) ** 2


def calc_instr_noise_sq(phot_params):
    """Combine all of the noise due to intrumentation into one value

    Parameters
    ----------
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        A PhotometricParameters object that carries details about the
        photometric response of the telescope.

    Returns
    -------
    inst_noise_sq : `float`
        The noise due to all of these sources added in quadrature in ADU counts
    """

    # instrumental squared noise in electrons
    inst_noise_sq = (
        phot_params.nexp * phot_params.readnoise**2
        + phot_params.darkcurrent * phot_params.exptime * phot_params.nexp
        + phot_params.nexp * phot_params.othernoise**2
    )

    # convert to ADU counts
    inst_noise_sq = inst_noise_sq / (phot_params.gain * phot_params.gain)

    return inst_noise_sq


def calc_total_non_source_noise_sq(sky_sed, hardwarebandpass, phot_params, fwhm_eff):
    """Calculate the noise due to instrumentation and sky background.

    Parameters
    ----------
    sky_sed : `rubin_sim.phot_utils.Sed`
        A Sed object representing the sky (normalized so that
        sky_sed.calc_mag() gives the sky brightness in
        magnitudes per square arcsecond)
    hardwarebandpass : `rubin_sim.phot_utils.Bandpass`
        A Bandpass object containing just the instrumentation
        throughputs (no atmosphere)
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        A PhotometricParameters object containing information
        about the photometric properties of the telescope.
    fwhm_eff : `float`
        fwhm_eff in arcseconds

    Returns
    -------
    total_noise_sq : `float`
        total non-source noise squared (in ADU counts)
        (this is simga^2_tot * neff in equation 41 of the SNR document
        https://ls.st/LSE-40 )
    """

    # Calculate the effective number of pixels for double-Gaussian PSF
    neff = calc_neff(fwhm_eff, phot_params.platescale)

    # Calculate the counts from the sky.
    # We multiply by two factors of the platescale because we expect the
    # sky_sed to be normalized such that calc_adu gives counts per
    # square arc second, and we need to convert to counts per pixel.

    skycounts = (
        sky_sed.calc_adu(hardwarebandpass, phot_params=phot_params)
        * phot_params.platescale
        * phot_params.platescale
    )

    # Calculate the square of the noise due to instrumental effects.
    # Include the readout noise as many times as there are exposures

    noise_instr_sq = calc_instr_noise_sq(phot_params=phot_params)

    # Calculate the square of the noise due to sky background poisson noise
    noise_sky_sq = skycounts / phot_params.gain

    # Discount error in sky measurement for now
    noise_skymeasurement_sq = 0

    total_noise_sq = neff * (noise_sky_sq + noise_instr_sq + noise_skymeasurement_sq)

    return total_noise_sq


def calc_sky_counts_per_pixel_for_m5(m5target, total_bandpass, phot_params, fwhm_eff=0.83):
    """Calculate the skycounts per pixel.

    Calculate the number of sky counts per pixel expected for a given
    value of the 5-sigma limiting magnitude (m5)

    Parameters
    ----------
    m5target : `float`
        The desired value of m5.
    total_bandpass : `rubin_sim.phot_utils.Bandpass`
        A bandpass object representing the total throughput of the telescope
        (instrumentation plus atmosphere).
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        A photometric parameters object containing the photometric response
        information for Rubin.
    fwhm_eff : `float`
        fwhm_eff in arcseconds. Default 0.83

    Returns
    -------
    sky_counts_target : `float`
        The expected number of sky counts per pixel (ADU/pixel).

    Notes
    -----
    The 5-sigma limiting magnitude (m5) for an observation is
    determined by a combination of the telescope and camera parameters
    (such as diameter of the mirrors and the readnoise) together with the
    sky background.
    """

    # instantiate a flat SED
    flat_sed = Sed()
    flat_sed.set_flat_sed()

    # normalize the SED so that it has a magnitude equal to the desired m5
    f_norm = flat_sed.calc_flux_norm(m5target, total_bandpass)
    flat_sed.multiply_flux_norm(f_norm)
    source_counts = flat_sed.calc_adu(total_bandpass, phot_params=phot_params)

    # calculate the effective number of pixels for a double-Gaussian PSF
    neff = calc_neff(fwhm_eff, phot_params.platescale)

    # calculate the square of the noise due to the instrument
    noise_instr_sq = calc_instr_noise_sq(phot_params=phot_params)

    # now solve equation 41 of the SNR document for the neff * sigma_total^2
    # term given snr=5 and counts as calculated above
    # SNR document can be found at
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40

    n_sigma_sq = (source_counts * source_counts) / 25.0 - source_counts / phot_params.gain

    sky_noise_target = n_sigma_sq / neff - noise_instr_sq
    sky_counts_target = sky_noise_target * phot_params.gain

    # TODO:
    # This method should throw an error if sky_counts_target is negative
    # unfortunately, that currently happens for default values of
    # m5 as taken from arXiv:0805.2366, table 2.  Adding the error
    # should probably wait for a later issue in which we hash out what
    # the units are for all of the parameters stored in PhotometricDefaults.

    return sky_counts_target


def calc_m5(skysed, total_bandpass, hardware, phot_params, fwhm_eff=0.83):
    """Calculate the AB magnitude of a 5-sigma source above sky background.

    Parameters
    ----------
    skysed : `rubin_sim.phot_utils.Sed`
        An SED representing the sky background emission, normalized such that
        skysed.calc_mag(Bandpass) returns the expected sky brightness in
        magnitudes per sq arcsecond.
    total_bandpass : `rubin_sim.phot_utils.Bandpass`
        The Bandpass representing the total throughput of the telescope
        (instrument plus atmosphere).
    hardware : `rubin_sim.phot_utils.Bandpass`
        The Bandpass representing the throughput of the telescope instrument
        only (no atmosphere).
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    fwhm_eff : `float`
        FWHM in arcseconds.

    Returns
    -------
    mag_5sigma : `float`
        The value of m5 for the given bandpass and sky SED

    Notes
    -----
    The 5-sigma limiting magnitude (m5) for an observation is determined by
    a combination of the telescope and camera parameters (such as diameter
    of the mirrors and the readnoise) together with the sky background. This
    method (calc_m5) calculates the expected m5 value for an observation given
    a sky background Sed and hardware parameters.

    This comes from equation 45 of the SNR document (v1.2, May 2010)
    https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40
    """

    # create a flat fnu source
    flatsource = Sed()
    flatsource.set_flat_sed()
    snr = 5.0
    v_n = calc_total_non_source_noise_sq(skysed, hardware, phot_params, fwhm_eff)

    counts_5sigma = (snr**2) / 2.0 / phot_params.gain + numpy.sqrt(
        (snr**4) / 4.0 / phot_params.gain + (snr**2) * v_n
    )

    # renormalize flatsource so that it has the required counts to be a
    # 5-sigma detection given the specified background
    counts_flat = flatsource.calc_adu(total_bandpass, phot_params=phot_params)
    flatsource.multiply_flux_norm(counts_5sigma / counts_flat)

    # Calculate the AB magnitude of this source.
    mag_5sigma = flatsource.calc_mag(total_bandpass)
    return mag_5sigma


def mag_error_from_snr(snr):
    """Convert flux signal to noise ratio to an error in magnitude.

    Parameters
    ----------
    snr : `float`
        The signal to noise ratio (a flux-related measurement).

    Returns
    -------
    mag_error : `float`
        Corresponding error in magnitude.
    """

    # see www.ucolick.org/~bolte/AY257/s_n.pdf section 3.1
    return 2.5 * numpy.log10(1.0 + 1.0 / snr)


def calc_gamma(bandpass, m5, phot_params):
    """Calculate gamma parameter.

    Calculate the gamma parameter used for determining photometric
    signal to noise in equation 5 of the LSST overview paper
    (arXiv:0805.2366)

    Parameters
    ----------
    bandpass : `Bandpass`
        Bandpass for which you desire to calculate the gamma parameter.
    m5 : `float`
        The magnitude of a 5-sigma point source detection.
    phot_params : `PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.

    Returns
    -------
    gamma : `float`
        The gamma value for this bandpass.
    """
    # This is based on the LSST SNR document (v1.2, May 2010)
    # https://docushare.lsstcorp.org/docushare/dsweb/ImageStoreViewer/LSE-40
    # as well as equations 4-6 of the overview paper (arXiv:0805.2366)

    # instantiate a flat SED
    flat_sed = Sed()
    flat_sed.set_flat_sed()

    # normalize the SED so that it has a magnitude equal to the desired m5
    f_norm = flat_sed.calc_flux_norm(m5, bandpass)
    flat_sed.multiply_flux_norm(f_norm)
    counts = flat_sed.calc_adu(bandpass, phot_params=phot_params)

    # The expression for gamma below comes from:
    #
    # 1) Take the approximation N^2 = N0^2 + alpha S from footnote 88
    # in the overview paper where N is the noise in flux of a source,
    # N0 is the noise in flux due to sky brightness and instrumentation,
    # S is the number of counts registered from the source and
    # alpha is some constant
    #
    # 2) Divide by S^2 and demand that N/S = 0.2 for a source detected at m5.
    # Solve the resulting equation for alpha in terms of N0 and S5
    # (the number of counts from a source at m5)
    #
    # 3) Substitute this expression for alpha into the equation for (N/S)^2
    # for a general source.  Re-factor the equation so that it looks like
    # equation 5 of the overview paper (note that x = S5/S).
    # This should give you gamma = (N0/S5)^2
    #
    # 4) Solve equation 41 of the SNR document for neff * sigma_total^2 term
    # given snr=5 and counts as calculated above.
    # Note that neff * sigma_total^2 is N0^2 in the equation above
    #
    # This should give you

    gamma = 0.04 - 1.0 / (counts * phot_params.gain)

    return gamma


def calc_snr_m5(magnitude, bandpass, m5, phot_params, gamma=None):
    """Calculate the SNR of a source based on the 5-sigma limit for an
    observation.

    Calculate signal to noise in flux using the model from equation (5)
    of arXiv:0805.2366

    Parameters
    ----------
    magnitude : `float` or `np.ndarray`, (N,)
        Magnitudes of the sources whose signal to noise you are calculating.
    bandpass : `rubin_sim.phot_utils.Bandpass`
        The Bandpass in which the magnitude was calculated
        (total instrument + atmosphere).
    m5 : `float`
        The 5-sigma point source limiting magnitude of the exposure.
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    gamma : `float`, opt
        The gamma parameter from equation(5) of arXiv:0805.2366.
        If not provided, this method will calculate it.

    Returns
    -------
    snr : `float` or `np.ndarray`, (N,)
        The SNR of the input magnitude.
    gamma : `float`
        The gamma parameter for the Bandpass.
    """

    if gamma is None:
        gamma = calc_gamma(bandpass, m5, phot_params=phot_params)

    dummy_sed = Sed()
    m5_flux = dummy_sed.flux_from_mag(m5)
    source_flux = dummy_sed.flux_from_mag(magnitude)

    flux_ratio = m5_flux / source_flux

    noise = numpy.sqrt((0.04 - gamma) * flux_ratio + gamma * flux_ratio * flux_ratio)

    return 1.0 / noise, gamma


def calc_mag_error_m5(magnitude, bandpass, m5, phot_params, gamma=None):
    """Calculate magnitude error using the model from equation (5)
    of arXiv:0805.2366

    Parameters
    ----------
    magnitude : `float`
        Magnitude of the source.
    bandpass : `rubin_sim.phot_utils.Bandpass`
        The Bandpass in which to calculate the magnitude error
        (total instrument + atmosphere).
    m5 : `float`
        The 5-sigma point source limiting magnitude.
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    gamma : `float`, optional
        The gamma parameter from equation(5) of arXiv:0805.2366.
        If not provided, this method will calculate it.

    Returns
    -------
    mag_error : `float` or `np.ndarray`, (N,)
        The magnitude error of the input magnitude.
    gamma : `float`
        The gamma parameter for the Bandpass.
    """

    snr, gamma = calc_snr_m5(magnitude, bandpass, m5, phot_params, gamma=gamma)

    if phot_params.sigma_sys is not None:
        return (
            numpy.sqrt(numpy.power(mag_error_from_snr(snr), 2) + numpy.power(phot_params.sigma_sys, 2)),
            gamma,
        )
    else:
        return mag_error_from_snr(snr), gamma


def calc_snr_sed(
    source_sed,
    total_bandpass,
    sky_sed,
    hardwarebandpass,
    phot_params,
    fwhm_eff,
    verbose=False,
):
    """Calculate the signal to noise ratio for a source, based on the SED.

    For a given source, sky sed, total bandpass and hardware bandpass,
    as well af fwhm_eff / exptime, calculates the SNR with
    optimal PSF extraction assuming a double-gaussian PSF.

    Parameters
    ----------
    source_sed : `rubin_sim.phot_utils.Sed`
        A SED representing the source, normalized such that source_sed.calc_mag
        gives the desired magnitude.
    total_bandpass : `rubin_sim.phot_utils.Bandpass`
        The Bandpass representing the total throughput of the telescope
        (instrument plus atmosphere).
    sky_sed : `rubin_sim.phot_utils.Sed`
        A SED representing the sky background emission, normalized such that
        skysed.calc_mag(Bandpass) returns the expected sky brightness in
        magnitudes per sq arcsecond.
    hardware : `rubin_sim.phot_utils.Bandpass`
        The Bandpass representing the throughput of the telescope instrument
        only (no atmosphere).
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    fwhm_eff : `float`
        FWHM in arcseconds.
    verbose : `bool`
        Flag as to whether to print output about SNR.

    Returns
    -------
    snr : `float`
        Calculated SNR.
    """

    # Calculate the counts from the source.
    sourcecounts = source_sed.calc_adu(total_bandpass, phot_params=phot_params)

    # Calculate the (square of the) noise due to signal poisson noise.
    noise_source_sq = sourcecounts / phot_params.gain

    non_source_noise_sq = calc_total_non_source_noise_sq(sky_sed, hardwarebandpass, phot_params, fwhm_eff)

    # Calculate total noise
    noise = numpy.sqrt(noise_source_sq + non_source_noise_sq)
    # Calculate the signal to noise ratio.
    snr = sourcecounts / noise
    if verbose:
        skycounts = sky_sed.calc_adu(hardwarebandpass, phot_params) * (phot_params.platescale**2)
        noise_sky_sq = skycounts / phot_params.gain
        neff = calc_neff(fwhm_eff, phot_params.platescale)
        noise_instr_sq = calc_instr_noise_sq(phot_params)

        print("For Nexp %.1f of time %.1f: " % (phot_params.nexp, phot_params.exptime))
        print("Counts from source: %.2f  Counts from sky: %.2f" % (sourcecounts, skycounts))
        print("fwhm_eff: %.2f('')  Neff pixels: %.3f(pix)" % (fwhm_eff, neff))
        print(
            "Noise from sky: %.2f Noise from instrument: %.2f"
            % (numpy.sqrt(noise_sky_sq), numpy.sqrt(noise_instr_sq))
        )
        print("Noise from source: %.2f" % (numpy.sqrt(noise_source_sq)))
        print(" Total Signal: %.2f   Total Noise: %.2f    SNR: %.2f" % (sourcecounts, noise, snr))
        # Return the signal to noise value.
    return snr


def calc_mag_error_sed(
    source_sed,
    total_bandpass,
    sky_sed,
    hardware_bandpass,
    phot_params,
    fwhm_eff,
    verbose=False,
):
    """Calculate the magnitudeError for a source, given the bandpass(es)
    and sky SED.

    For a given source, sky sed, total bandpass and hardware bandpass, and
    fwhm_eff / exptime, calculates the SNR with optimal PSF extraction
    assuming a double-gaussian PSF.

    Parameters
    ----------
    source_sed : `rubin_sim.phot_utils.Sed`
        A SED representing the source, normalized such that source_sed.calc_mag
        gives the desired magnitude.
    total_bandpass : `rubin_sim.phot_utils.Bandpass`
        The Bandpass representing the total throughput of the telescope
        (instrument plus atmosphere).
    sky_sed : `rubin_sim.phot_utils.Sed`
        A SED representing the sky background emission, normalized such that
        skysed.calc_mag(Bandpass) returns the expected sky brightness in
        magnitudes per sq arcsecond.
    hardware_bandpass : `rubin_sim.phot_utils.Bandpass`
        The Bandpass representing the throughput of the telescope instrument
        only (no atmosphere).
    phot_params : `rubin_sim.phot_utils.PhotometricParameters`
        The PhotometricParameters class that carries details about the
        photometric response of the telescope.
    fwhm_eff : `float`
        FWHM in arcseconds.
    verbose : `bool`, optional
        Flag as to whether to print output about SNR.

    Returns
    -------
    mag_err : `float`
        Magnitude error in expected magnitude.
    """

    snr = calc_snr_sed(
        source_sed,
        total_bandpass,
        sky_sed,
        hardware_bandpass,
        phot_params,
        fwhm_eff,
        verbose=verbose,
    )

    if phot_params.sigma_sys is not None:
        return numpy.sqrt(numpy.power(mag_error_from_snr(snr), 2) + numpy.power(phot_params.sigma_sys, 2))
    else:
        return mag_error_from_snr(snr)


def calc_astrometric_error(mag, m5, fwhm_geom=0.7, nvisit=1, systematic_floor=10):
    """Calculate an expected astrometric error.

    The astrometric error can be estimated for catalog purposes by using the
    typical FWHM and n_visit = total number of visits, or can be used for a
    single visit by using the actual FWHM and n_visit =1.


    Parameters
    ----------
    mag: `float`
        Magnitude of the source
    m5: `float`
        Point source five sigma limiting magnitude of the image
        (or typical depth per image).
    fwhm_geom: `float`, optional
        The geometric (physical) FWHM of the image, in arcseconds.
    nvisit: `int`, optional
        The number of visits/measurement. Default 1.
        If this is >1, the random error contribution is reduced by
        sqrt(nvisits).
    systematic_floor: `float`, optional
        The systematic noise floor for the astrometric measurements,
        in mas. Default 10mas.

    Returns
    -------
    astrom_err : `float`
        Astrometric error for a given SNR, in mas.
    """
    # The astrometric error can be applied to parallax or proper motion
    # (for n_visit>1).
    # If applying to proper motion, should also divide by the # of years
    # of the survey.
    # This is also referenced in the astroph/0805.2366 paper.
    # D. Monet suggests sqrt(Nvisit/2) for first 3 years, sqrt(N) for longer,
    # in reduction of error.
    # Zeljko says 'be conservative', so removing this reduction for now.
    rgamma = 0.039
    xval = numpy.power(10, 0.4 * (mag - m5))
    # The average fwhm_eff is 0.7" (or 700 mas), but user can specify.
    # Convert to mas.
    seeing = fwhm_geom * 1000.0
    error_rand = seeing * numpy.sqrt((0.04 - rgamma) * xval + rgamma * xval * xval)
    error_rand = error_rand / numpy.sqrt(nvisit)
    # The systematic error floor in astrometry (mas).
    error_sys = systematic_floor
    astrom_error = numpy.sqrt(error_sys * error_sys + error_rand * error_rand)
    return astrom_error


def scale_sky_m5(m5target, skysed, total_bandpass, hardware, phot_params, fwhm_eff=0.83):
    """
    Take an SED representing the sky and normalize it so that
    m5 (the magnitude at which an object is detected in this
    bandpass at 5-sigma) is set to some specified value.

    The 5-sigma limiting magnitude (m5) for an observation is
    determined by a combination of the telescope and camera parameters
    (such as diameter of the mirrors and the readnoise) together with the
    sky background. This method (set_m5) scales a provided sky background
    Sed so that an observation would have a target m5 value, for the
    provided hardware parameters. Using the resulting Sed in the
    'calcM5' method will return this target value for m5.

    Note that the returned SED will be renormalized such that calling the
    method self.calcADU(hardwareBandpass) on it will yield the number of
    counts per square arcsecond in a given bandpass.
    """

    # This is based on the LSST SNR document (v1.2, May 2010)
    # www.astro.washington.edu/users/ivezic/Astr511/LSST_SNRdoc.pdf

    sky_counts_target = calc_sky_counts_per_pixel_for_m5(
        m5target, total_bandpass, fwhm_eff=fwhm_eff, phot_params=phot_params
    )

    sky_sed_out = Sed(wavelen=numpy.copy(skysed.wavelen), flambda=numpy.copy(skysed.flambda))

    sky_counts = (
        sky_sed_out.calc_adu(hardware, phot_params=phot_params)
        * phot_params.platescale
        * phot_params.platescale
    )
    sky_sed_out.multiply_flux_norm(sky_counts_target / sky_counts)

    return sky_sed_out



from .agn_time_lag_metric import *
from .agnstructure import *
from .area_summary_metrics import *
from .base_metric import *
from .brown_dwarf_metric import *
from .cadence_metrics import *
from .calibration_metrics import *
from .chip_vendor_metric import *
from .color_slope_metrics import *
from .coverage_metric import *
from .crowding_metric import *
from .cumulative_metric import *
from .dcr_metric import *
from .exgal_m5 import *
from .fft_metric import *
from .galactic_plane_metrics import *
from .galplane_time_sampling_metrics import *
from .hourglass_metric import *
from .incremental_template_metric import *
from .kuiper_metrics import *
from .mo_metrics import *
from .mo_summary_metrics import *
from .night_pointing_metric import *
from .optimal_m5_metric import *
from .pair_metric import *
from .periodic_detect_metric import *
from .phase_gap_metric import *
from .qso_number_counts_metric import *
from .scaling_metrics import *
from .schedview_metrics import *
from .season_metrics import *
from .simple_metrics import *
from .sky_sat_metric import *
from .sn_cadence_metric import *
from .sn_n_sn_metric import *
from .sn_sl_metric import *
from .sn_snr_metric import *
from .snr_weight import *
from .star_density import *
from .string_count_metric import *
from .summary_metrics import *
from .surfb_metric import *
from .technical_metrics import *
from .tgaps import *
from .transient_metrics import *
from .use_metrics import *
from .vector_metrics import *
from .visit_groups_metric import *
from .weak_lensing_systematics_metric import *



__all__ = ("GalaxyCountsMetric",)

import healpy as hp
import numpy as np
import scipy

from rubin_sim.maf.metrics import BaseMetric, ExgalM5


class GalaxyCountsMetric(BaseMetric):
    """Estimate the number of galaxies expected at a particular (extragalactic)
    coadded depth.
    """

    def __init__(self, m5_col="fiveSigmaDepth", nside=128, metric_name="GalaxyCounts", **kwargs):
        self.m5_col = m5_col
        super(GalaxyCountsMetric, self).__init__(col=self.m5_col, metric_name=metric_name, **kwargs)
        # Use the extinction corrected coadded depth metric to calculate
        # the depth at each point.
        self.coaddmetric = ExgalM5(m5_col=self.m5_col)
        # Total of 41253.0 galaxies across the sky (at what magnitude?).
        # This didn't seem to work quite right for me..
        self.scale = 41253.0 / hp.nside2npix(nside) / 5000.0
        # Reset units (otherwise uses magnitudes).
        self.units = "Galaxy Counts"

    def _gal_count(self, apparent_mag, coaddm5):
        # Order for galCount must be apparent mag, then coaddm5,
        # for scipy.integrate method.
        dn_gal = np.power(10.0, -3.52) * np.power(10.0, 0.34 * apparent_mag)
        completeness = 0.5 * scipy.special.erfc(apparent_mag - coaddm5)
        return dn_gal * completeness

    def run(self, data_slice, slice_point=None):
        # Calculate the coadded depth.
        coaddm5 = self.coaddmetric.run(data_slice)
        # Calculate the number of galaxies.
        # From Carroll et al, 2014 SPIE (http://arxiv.org/abs/1501.04733)
        # I'm not entirely certain this gives a properly calibrated number
        # of galaxy counts, however it is proportional to the expected number
        # at least (and should be within an order of magnitude)
        num_gal, int_err = scipy.integrate.quad(self._gal_count, -np.inf, 32, args=coaddm5)
        num_gal *= self.scale
        return num_gal



__all__ = ("make_bundle_list", "maf_night_report")

import argparse

import matplotlib

# Set matplotlib backend (to create plots where DISPLAY is not set).
matplotlib.use("Agg")
from . import db as db
from . import metricBundles as metricBundles
from . import metrics as metrics
from . import plots as plots
from . import slicers as slicers
from . import utils as utils
from .batches import col_map_dict


def make_bundle_list(
    db_file,
    night=1,
    nside=64,
    latCol="fieldDec",
    lonCol="fieldRA",
    notes=True,
    colmap=None,
):
    """
    Make a bundleList of things to run
    """

    if colmap is None:
        colmap = col_map_dict("opsimV4")

    mjdCol = "observationStartMJD"
    altCol = "altitude"
    azCol = "azimuth"
    # Construct sql queries for each filter and all filters
    filters = ["u", "g", "r", "i", "z", "y"]
    sqls = ['night=%i and filter="%s"' % (night, f) for f in filters]
    sqls.append("night=%i" % night)

    bundleList = []
    plotFuncs_lam = [plots.LambertSkyMap()]

    # Hourglass
    hourslicer = slicers.HourglassSlicer()
    display_dict = {"group": "Hourglass"}
    md = ""
    sql = "night=%i" % night
    metric = metrics.HourglassMetric(
        night_col=colmap["night"], mjd_col=colmap["mjd"], metric_name="Hourglass"
    )
    bundle = metricBundles.MetricBundle(
        metric, hourslicer, constraint=sql, info_label=md, display_dict=display_dict
    )
    bundleList.append(bundle)

    reg_slicer = slicers.HealpixSlicer(nside=nside, lon_col=lonCol, lat_col=latCol, lat_lon_deg=True)
    altaz_slicer = slicers.HealpixSlicer(
        nside=nside, lat_col=altCol, lat_lon_deg=True, lon_col=azCol, use_cache=False
    )

    unislicer = slicers.UniSlicer()
    for sql in sqls:
        # Number of exposures
        metric = metrics.CountMetric(mjdCol, metric_name="N visits")
        bundle = metricBundles.MetricBundle(metric, reg_slicer, sql)
        bundleList.append(bundle)
        metric = metrics.CountMetric(mjdCol, metric_name="N visits alt az")
        bundle = metricBundles.MetricBundle(metric, altaz_slicer, sql, plot_funcs=plotFuncs_lam)
        bundleList.append(bundle)

        metric = metrics.MeanMetric(mjdCol, metric_name="Mean Visit Time")
        bundle = metricBundles.MetricBundle(metric, reg_slicer, sql)
        bundleList.append(bundle)
        metric = metrics.MeanMetric(mjdCol, metric_name="Mean Visit Time alt az")
        bundle = metricBundles.MetricBundle(metric, altaz_slicer, sql, plot_funcs=plotFuncs_lam)
        bundleList.append(bundle)

        metric = metrics.CountMetric(mjdCol, metric_name="N_visits")
        bundle = metricBundles.MetricBundle(metric, unislicer, sql)
        bundleList.append(bundle)

        # Need pairs in window to get a map of how well it gathered SS pairs.

    # Moon phase.

    metric = metrics.NChangesMetric(col="filter", metric_name="Filter Changes")
    bundle = metricBundles.MetricBundle(metric, unislicer, "night=%i" % night)
    bundleList.append(bundle)

    metric = metrics.BruteOSFMetric()
    bundle = metricBundles.MetricBundle(metric, unislicer, "night=%i" % night)
    bundleList.append(bundle)

    metric = metrics.MeanMetric("slewTime")
    bundle = metricBundles.MetricBundle(metric, unislicer, "night=%i" % night)
    bundleList.append(bundle)

    metric = metrics.MinMetric("slewTime")
    bundle = metricBundles.MetricBundle(metric, unislicer, "night=%i" % night)
    bundleList.append(bundle)

    metric = metrics.MaxMetric("slewTime")
    bundle = metricBundles.MetricBundle(metric, unislicer, "night=%i" % night)
    bundleList.append(bundle)

    # Make plots of the solar system pairs that were taken in the night
    metric = metrics.PairMetric(mjd_col=mjdCol)
    sql = 'night=%i and (filter ="r" or filter="g" or filter="i")' % night
    bundle = metricBundles.MetricBundle(metric, reg_slicer, sql)
    bundleList.append(bundle)

    metric = metrics.PairMetric(mjd_col=mjdCol, metric_name="z Pairs")
    sql = 'night=%i and filter="z"' % night
    bundle = metricBundles.MetricBundle(metric, reg_slicer, sql)
    bundleList.append(bundle)

    # Plot up each visit
    metric = metrics.NightPointingMetric(mjd_col=mjdCol)
    slicer = slicers.UniSlicer()
    sql = "night=%i" % night
    plot_funcs = [plots.NightPointingPlotter()]
    bundle = metricBundles.MetricBundle(metric, slicer, sql, plot_funcs=plot_funcs)
    bundleList.append(bundle)

    # stats from the note column
    if notes:
        display_dict = {"group": "Basic Stats", "subgroup": "Percent stats"}
        metric = metrics.StringCountMetric(col="note", percent=True, metric_name="Percents")
        bundle = metricBundles.MetricBundle(metric, unislicer, sql, display_dict=display_dict)
        bundleList.append(bundle)
        display_dict["subgroup"] = "Count Stats"
        metric = metrics.StringCountMetric(col="note", metric_name="Counts")
        bundle = metricBundles.MetricBundle(metric, unislicer, sql, display_dict=display_dict)
        bundleList.append(bundle)

    return metricBundles.make_bundles_dict_from_list(bundleList)


def maf_night_report():
    """Generate a report on a single night."""

    parser = argparse.ArgumentParser(description="Python script to generate a report on a single night.")
    parser.add_argument(
        "db_file",
        type=str,
        default=None,
        help="full file path to the simulation sqlite file",
    )
    parser.add_argument(
        "--out_dir",
        type=str,
        default="night_report",
        help="Output directory for MAF outputs." + ' Default "Out"',
    )
    parser.add_argument(
        "--nside",
        type=int,
        default=64,
        help="Resolution to run Healpix grid at (must be 2^x). Default 64.",
    )
    parser.add_argument(
        "--lon_col",
        type=str,
        default="fieldRA",
        help="Column to use for RA values (can be a stacker dither column)." + " Default=fieldRA.",
    )
    parser.add_argument(
        "--lat_col",
        type=str,
        default="fieldDec",
        help="Column to use for Dec values (can be a stacker dither column)." + " Default=fieldDec.",
    )
    parser.add_argument("--night", type=int, default=1)
    parser.add_argument("--run_name", type=str, default="run_name")

    parser.set_defaults()
    args, extras = parser.parse_known_args()

    bundle_dict = make_bundle_list(
        args.db_file,
        nside=args.nside,
        lonCol=args.lon_col,
        latCol=args.lat_col,
        night=args.night,
    )

    for key in bundle_dict:
        bundle_dict[key].set_run_name(args.run_name)

    # Set up / connect to results_db.
    results_db = db.ResultsDb(out_dir=args.out_dir)

    # Set up metricBundleGroup.
    group = metricBundles.MetricBundleGroup(
        bundle_dict, args.db_file, out_dir=args.out_dir, results_db=results_db
    )
    group.run_all()
    group.plot_all()



import numpy as np

from .base_metric import BaseMetric

__all__ = ["TemplateTime"]


class TemplateTime(BaseMetric):
    """Find the time at which we expect to hit incremental template
    availability.

    Note that there are some complications to real template generation
    that make this an approximation and not an exact answer -- one aspect is
    that templates are generated in `patches` and not per pixel. However, it
    may be possible to generate parts of these patches at about 5 arcsecond
    scales, which implies running with a healpix slicer at nside=512 or 1024.

    Parameters
    ----------
    n_images_in_template : `int`, opt
        Number of qualified visits required for template generation.
        Default 3.
    seeing_percentile : `float`, opt
        Maximum percentile seeing to allow in the qualified images (0 - 100).

        Default 50.
    m5_percentile : `float`, opt
        Maximum percentile m5 to allow in the qualified images (0 - 100).

        Default 50.
    seeing_col : `str`, opt
        Name of the seeing column to use.
    m5_col : `str`, opt
        Name of the five sigma depth columns.
    night_col : `str`, opt
        Name of the column describing the night of the visit.
    mjd_col : `str`, opt
        Name of column describing time of the visit
    filter_col : `str`, opt
        Name of column describing filter
    """

    def __init__(
        self,
        n_images_in_template=3,
        seeing_percentile=50,
        m5_percentile=50,
        seeing_col="seeingFwhmEff",
        m5_col="fiveSigmaDepth",
        night_col="night",
        mjd_col="observationStartMJD",
        filter_col="filter",
        **kwargs,
    ):
        self.n_images_in_template = n_images_in_template
        self.seeing_percentile = seeing_percentile
        self.m5_percentile = m5_percentile
        self.seeing_col = seeing_col
        self.m5_col = m5_col
        self.night_col = night_col
        self.mjd_col = mjd_col
        self.filter_col = filter_col
        if "metric_name" in kwargs:
            self.metric_name = kwargs["metric_name"]
            del kwargs["metric_name"]
        else:
            self.metric_name = "TemplateTime"
        super().__init__(
            col=[self.seeing_col, self.m5_col, self.night_col, self.mjd_col, self.filter_col],
            metric_name=self.metric_name,
            units="days",
            **kwargs,
        )

    def run(self, data_slice, slice_point=None):
        result = {}

        # Bail if not enough visits at all
        if len(data_slice) < self.n_images_in_template:
            return self.badval

        # Check that the visits are sorted in time
        data_slice.sort(order=self.mjd_col)

        # Find the threshold seeing and m5
        bench_seeing = np.percentile(data_slice[self.seeing_col], self.seeing_percentile)
        bench_m5 = np.percentile(data_slice[self.m5_col], 100 - self.m5_percentile)

        seeing_ok = np.where(data_slice[self.seeing_col] < bench_seeing, True, False)

        m5_ok = np.where(data_slice[self.m5_col] > bench_m5, True, False)

        ok_template_input = np.where(seeing_ok & m5_ok)[0]
        if (
            len(ok_template_input) < self.n_images_in_template
        ):  # If seeing_ok and/or m5_ok are "false", returned as bad value
            return self.badval

        idx_template_created = ok_template_input[
            self.n_images_in_template - 1
        ]  # Last image needed for template
        idx_template_inputs = ok_template_input[: self.n_images_in_template]  # Images included in template

        Night_template_created = data_slice[self.night_col][idx_template_created]
        N_nights_without_template = Night_template_created - data_slice[self.night_col][0]

        where_template = (
            data_slice[self.night_col] > Night_template_created
        )  # of later images where we have a template
        N_images_with_template = np.sum(where_template)

        template_m5 = 1.25 * np.log10(np.sum(10.0 ** (0.8 * data_slice[self.m5_col][idx_template_inputs])))

        # derive variance-weighted PSF width
        # 1. normalize the weights
        template_f0 = np.sqrt(np.sum(25 * 10 ** (0.8 * data_slice[self.m5_col][idx_template_inputs])))
        # 2. compute per-input noise
        image_noise = template_f0 / 5 * 10 ** (-0.4 * data_slice[self.m5_col])
        # 3. used variance-weighted sum of squares to derive template seeing
        template_seeing = np.sqrt(
            np.sum(
                (data_slice[self.seeing_col][idx_template_inputs] / image_noise[idx_template_inputs]) ** 2.0
            )
        )

        delta_m5 = -2.5 * np.log10(np.sqrt(1.0 + 10 ** (-0.8 * (template_m5 - data_slice[self.m5_col]))))
        diff_m5s = data_slice[self.m5_col] + delta_m5

        n_alerts_per_diffim = 1e4 * 10 ** (0.6 * (diff_m5s - 24.7))

        result["Night_template_created"] = Night_template_created
        result["N_nights_without_template"] = N_nights_without_template
        result["N_images_until_template"] = idx_template_created + 1
        result["N_images_with_template"] = N_images_with_template
        result["Template_m5"] = template_m5
        result["Template_seeing"] = template_seeing
        result["Total_alerts"] = np.sum(n_alerts_per_diffim[where_template])
        result["Template_input_m5s"] = data_slice[self.m5_col][idx_template_inputs]
        result["Template_input_seeing"] = data_slice[self.seeing_col][idx_template_inputs]
        result["Fraction_better_template_seeing"] = np.sum(
            (data_slice[self.seeing_col][where_template] > template_seeing)
        ) / np.sum(where_template)
        result["Diffim_lc"] = {
            "mjd": data_slice[self.mjd_col][where_template],
            "night": data_slice[self.night_col][where_template],
            "band": data_slice[self.filter_col][where_template],
            "diff_m5": diff_m5s[where_template],
            "science_m5": data_slice[self.m5_col][where_template],
            "template_m5": template_m5 * np.ones(np.sum(where_template)),
            "science_seeing": data_slice[self.seeing_col][where_template],
            "template_seeing": template_seeing * np.ones(np.sum(where_template)),
            "n_alerts": n_alerts_per_diffim[where_template],
        }

        return result

    def reduce_Night_template_created(self, metric_val):  # returns night of template creation
        return metric_val["Night_template_created"]

    def reduce_N_nights_without_template(
        self, metric_val
    ):  # returns number of nights needed to complete template
        return metric_val["N_nights_without_template"]

    def reduce_N_images_until_template(
        self, metric_val
    ):  # returns number of images needed to complete template
        return metric_val["N_images_until_template"]

    def reduce_N_images_with_template(self, metric_val):  # returns number of images with a template
        return metric_val["N_images_with_template"]

    def reduce_Template_m5(self, metric_val):  # calculated coadded m5 of resulting template
        return metric_val["Template_m5"]

    def reduce_Template_seeing(self, metric_val):  # calculated seeing of resulting template
        return metric_val["Template_seeing"]

    def reduce_fraction_better_template_seeing(self, metric_val):  # calculated seeing of resulting template
        return metric_val["Fraction_better_template_seeing"]

    def reduce_Total_alerts(self, metric_val):
        return metric_val["Total_alerts"]
